# -*- coding: utf-8 -*-
import numpy as np
import numba as nb
import warnings


class RebinError(Exception):
    """Exception raised by rebin operations."""

    pass


class RebinWarning(UserWarning):
    """Warnings displayed by rebin operations."""

    pass


def _check_ndim(arr, ndim, arr_name='array'):
    """Check the dimensionality of a numpy array

    Check that the array arr has dimension ndim.

    Args:
      arr: numpy array for which dimensionality is checked
      ndim: a scalar number or an iterable
      arr_name: name of array, just for use in the AssertionError message

    Raises:
      AssertionError: if arr does not have dimension ndim
    """
    if (arr.ndim != ndim) and (arr.ndim not in ndim):
        raise RebinError('{}({}) is not {}D'.format(
            arr_name, arr.shape, ndim))


def _check_monotonic_increasing(arr, arr_name='array'):
    """Check that a numpy array is monotonically increasing

    Args:
      arr: numpy array for checking
      arr_name: name of array, just for use in the AssertionError message

    Raises:
      AssertionError: if arr is not monotonically increasing
    """
    # Check that elements along the last axis are increasing or
    # neighboring elements are equal
    tmp = np.diff(arr)
    if not np.all((tmp > 0) | np.isclose(tmp, 0)):
        raise RebinError('{} is not monotonically increasing: {}'.format(
            arr_name, arr))


def _check_partial_overlap(in_edges, out_edges):
    """
    Args:
        in_edges (np.ndarray): an array of the input bin edges (1D or 2D)
            [num_spectra, num_bins_in + 1] or [num_bins_in + 1]
        out_edges (np.ndarray): an array of the output bin edges
            [num_bins_out]

    Raises:
        RebinWarning: for the following cases:

            old:   └┴┴ ...
            new: └┴┴┴┴ ...

            old: ... ┴┴┴┘
            new: ... ┴┴┴┴┴┘

    """
    if np.any(in_edges[..., 0] > out_edges[0]):
        warnings.warn(
            'The first input edge is larger than the first output edge, ' +
            'zeros will padded on the left side of the new spectrum',
            RebinWarning)
    if np.any(in_edges[..., -1] < out_edges[-1]):
        warnings.warn(
            'The last input edge is smaller than the last output edge, ' +
            'zeros will padded on the right side of the new spectrum',
            RebinWarning)


def _check_any_overlap(in_edges, out_edges):
    """
    Args:
        in_edges (np.ndarray): an array of the input bin edges (1D or 2D)
            [num_spectra, num_bins_in + 1] or [num_bins_in + 1]
        out_edges (np.ndarray): an array of the output bin edges
            [num_bins_out]

    Raises:

        RebinError: for the following cases:

            old:   └┴┴┴┴┘
            new:             └┴┴┴┘

            old:           └┴┴┴┴┘
            new:   └┴┴┴┘
    """
    if np.any(in_edges[..., -1] <= out_edges[0]):
        raise RebinError('Input edges are all smaller than output edges')
    if np.any(in_edges[..., 0] >= out_edges[-1]):
        raise RebinError('Input edges are all larger than output edges')


def _broadcast(arr, shape):
    """
    broadcast arr out to the first dimension in shape
    specifically for the case 1D -> 2D
    copy req'd: the readonly array doesn't work w/ numba
    """
    if (arr.ndim == 1) and (len(shape) == 2):
        return np.copy(np.broadcast_to(arr, (shape[0], arr.shape[0])))
    else:
        return arr


def _check_shape(arr0, arr1, arr0_name='array0', arr1_name='array1',
                 edges=False):
    arr0_shape = arr0.shape
    arr1_shape = arr1.shape
    if edges:
        if len(arr1_shape) == 1:
            arr1_shape = (arr1_shape[0] - 1,)
        else:
            arr1_shape = (arr1_shape[0], arr1_shape[1] - 1,)
    if arr0_shape != arr1_shape:
        raise RebinError(
            '{}{} does not have a shape compatible with {}{}'.format(
                arr0_name, arr0_shape, arr1_name, arr1_shape))


@nb.jit(nb.f8(nb.f8, nb.f8, nb.f8, nb.f8), nopython=True)
def _linear_offset(slope, cts, low, high):
    """
    Calculate the offset of the linear aproximation of slope when splitting
    counts between bins.

    Args:
      slope:
      cts: counts within the bin
      low: lower bin edge energy
      high: higher bin edge energy

    Returns:
      the offset
    """
    if np.abs(slope) < 1e-6:
        offset = cts / (high - low)
    else:
        offset = (cts - slope / 2. * (high**2 - low**2)) / (high - low)
    return offset


@nb.jit(nb.f8(nb.f8, nb.f8, nb.f8), nopython=True)
def _slope_integral(x, m, b):
    '''
    The indefinite integral of y = mx + b, with an x value substituted in.

    Args:
      x: the x-value
      m: the value of the slope
      b: the y-offset value

    Returns:
      The indefinite integral of y = mx + b, with an x value substituted in.
      (m x^2 / 2 + b x)
    '''
    return m * x**2 / 2 + b * x


@nb.jit(nb.f8(nb.f8, nb.f8, nb.f8, nb.f8), nopython=True)
def _counts(m, b, x_low, x_high):
    '''
    the definite integral of y = mx + b

    Computes the area under a linear approximation of the changing count
    rate in the vincity of relevant bins.  Edges of this integration
    are low and high while offset is provided from _linear_offset and cts
    from the bin being partitioned.

    Args:
      m: the value of slope m
      b: the value of y-offset b
      x_low: the "high" value for x, to be substituted into the integral
      x_high: the "low" value for x, to be substituted into the integral

    Returns:
      the definite integral of y = mx + b
    '''
    return _slope_integral(x_high, m, b) - _slope_integral(x_low, m, b)


@nb.jit(nb.f8[:](nb.f8[:], nb.f8[:], nb.f8[:], nb.f8[:]),
        locals={'in_idx': nb.u4, 'out_idx': nb.u4, 'counts': nb.f8,
                'slope': nb.f8, 'offset': nb.f8, 'low': nb.f8, 'high': nb.f8},
        nopython=True)
def _rebin_interpolation(in_spectrum, in_edges, out_edges, slopes):
    """
    Rebins a spectrum using linear interpolation.

    Keeps a running counter of two loop indices: in_idx & out_idx

    Args:
      in_spectrum: iterable of input spectrum counts in each bin
      in_edges: iterable of input bin edges (len = len(in_spectrum) + 1)
      out_edges: iterable of output bin edges
      slopes: the slopes of each histogram bin, with the lines drawn between
        each bin edge. (len = len(in_spectrum))

    Returns:
      1D numpy array of rebinned spectrum counts in each bin
    """
    out_spectrum = np.zeros(out_edges.shape[0] - 1)  # init output
    # in_idx: input bin or left edge
    #         init to the first in_bin which overlaps the 0th out_bin
    in_idx = max(0, np.searchsorted(in_edges, out_edges[0]) - 1)
    # Under-flow handling: Put all counts from in_bins that are completely to
    # the left of the 0th out_bin into the 0th out_bin
    out_spectrum[0] += np.sum(in_spectrum[:in_idx])
    # out_idx: output bin or left edge
    #          init to the first out_bin which overlaps the 0th in_bin
    out_idx = max(0, np.searchsorted(out_edges, in_edges[0]) - 1)
    # loop through input bins (starting from the above in_idx value):
    for in_idx in range(in_idx, len(in_spectrum)):
        # input bin info/data
        in_left_edge = in_edges[in_idx]
        in_right_edge = in_edges[in_idx + 1]
        counts = in_spectrum[in_idx]
        slope = slopes[in_idx]
        offset = _linear_offset(slope, counts, in_left_edge, in_right_edge)
        # loop through output bins that overlap with the current input bin
        for out_idx in range(out_idx, len(out_spectrum)):
            out_left_edge = out_edges[out_idx]
            out_right_edge = out_edges[out_idx + 1]
            if out_left_edge > in_right_edge:
                out_idx -= 1  # rewind back to previous out_bin; not done yet
                break  # break out of out_bins loop, move on to next in_bin
            # Low edge for interpolation
            if out_idx == 0:
                # under-flow handling:
                # all counts in this in_bin goes into the 0th out_bin
                low = in_left_edge  # == min(in_left_edge, out_edges[0])
            else:
                low = max(in_left_edge, out_left_edge)
            # High edge for interpolation
            if out_idx == len(out_spectrum) - 1:
                # over-flow handling:
                # all counts in this in_bin goes into this out_bin
                high = in_right_edge
            else:
                high = min(in_right_edge, out_right_edge)
            # Calc counts for this bin
            out_spectrum[out_idx] += _counts(slope, offset, low, high)
    return out_spectrum


@nb.jit(nb.i8[:](nb.i8[:], nb.f8[:], nb.f8[:], nb.f8[:]),
        locals={'energies': nb.f8[:], 'in_idx': nb.u4, 'energy_idx': nb.u4},
        nopython=True)
def _rebin_listmode(in_spectrum, in_edges, out_edges, slopes):
    """
    Stochastic rebinning method: spectrum-histogram to listmode then back

    TODO Assume piecewise constant (ie steps, flat within each bin)
         distribution for in_spectrum (for now). slopes is unused.

    Args:
      in_spectrum: iterable of input spectrum counts in each bin
      in_edges: iterable of input bin edges (len = len(in_spectrum) + 1)
      out_edges: iterable of output bin edges
      slopes: unused, just for keeping number of arguments the same as
              _rebin_interpolation()

    Returns:
      1D numpy array of rebinned spectrum counts in each bin
    """
    energies = np.zeros(np.sum(in_spectrum))
    energy_idx_start = 0
    # loop through input bins:
    # in_idx = index of in_bin or in_left_edge
    for in_idx in range(len(in_spectrum)):
        bin_counts = in_spectrum[in_idx]
        energy_idx_stop = energy_idx_start + bin_counts
        in_left_edge = in_edges[in_idx]
        in_right_edge = in_edges[in_idx + 1]
        energies[energy_idx_start:energy_idx_stop] = in_left_edge + (
            in_right_edge - in_left_edge) * np.random.rand(bin_counts)
        energy_idx_start = energy_idx_stop
    # bin the energies (drops the energies outside of the out-binning-range)
    out_spectrum = np.histogram(energies, bins=out_edges)[0]
    # add the under/flow counts back into the out_spectrum
    out_spectrum[0] += (energies < out_edges[0]).nonzero()[0].size
    out_spectrum[-1] += (energies > out_edges[-1]).nonzero()[0].size
    return out_spectrum


@nb.jit(nb.f8[:, :](nb.f8[:, :], nb.f8[:, :], nb.f8[:], nb.f8[:, :]),
        locals={'i': nb.u4}, nopython=True)
def _rebin2d_interpolation(in_spectra, in_edges, out_edges, slopes):
    """
    Rebins a 2D array of spectra using linear interpolation

    N.B. Does not keep Poisson statistics

    Wrapper around _rebin_interpolation (1D) for the 2D case

    Args:
      in_spectra: np.2darray, shape (num_spectra, num_bins_in)
        array of input spectrum counts of shape
      in_edges: np.2darray, shape (num_spectra, num_bins_in + 1)
        array of the input bin edges of shape
      out_edges: np.1darray
        array of the output bin edges
      slopes:

    Returns:
      np.2darray, shape (num_spectra, num_bins_in)
      The rebinned spectra
    """
    # Init output
    out_spectra = np.zeros((in_spectra.shape[0], out_edges.shape[0] - 1))
    for i in np.arange(in_spectra.shape[0]):
        out_spectra[i, :] = _rebin_interpolation(
            in_spectra[i, :], in_edges[i, :], out_edges, slopes[i, :])
    return out_spectra


@nb.jit(nb.i8[:, :](nb.i8[:, :], nb.f8[:, :], nb.f8[:], nb.f8[:, :]),
        locals={'i': nb.u4}, nopython=True)
def _rebin2d_listmode(in_spectra, in_edges, out_edges, slopes):
    """
    Rebins a 2D array of spectra stochastically: histogram to listmode and back

    Wrapper around _rebin_listmode (1D) for the 2D case

    Args:
      in_spectra: np.2darray, shape (num_spectra, num_bins_in)
        array of input spectrum counts of shape
      in_edges: np.2darray, shape (num_spectra, num_bins_in + 1)
        array of the input bin edges of shape
      out_edges: np.1darray
        array of the output bin edges
      slopes: TODO unused, just for keeping number of arguments the same as
              _rebin_interpolation()

    Returns:
      np.2darray, shape (num_spectra, num_bins_in)
      The rebinned spectra
    """
    # Init output
    out_spectra = np.zeros((in_spectra.shape[0], out_edges.shape[0] - 1),
                           np.int64)
    for i in np.arange(in_spectra.shape[0]):
        out_spectra[i, :] = _rebin_listmode(in_spectra[i, :], in_edges[i, :],
                                            out_edges, slopes[i, :])
    return out_spectra


def rebin(in_spectra, in_edges, out_edges, method="interpolation",
          slopes=None, zero_pad_warnings=True):
    """
    Spectra rebinning via deterministic or stochastic methods.

    Args:
        in_spectrum (np.ndarray): an array of input spectrum counts (1D or 2D)
            [num_spectra, num_bins_in] or [num_bins_in]
        in_edges (np.ndarray): an array of the input bin edges (1D or 2D)
            [num_spectra, num_bins_in + 1] or [num_bins_in + 1]
        out_edges (np.ndarray): an array of the output bin edges
            [num_bins_out]
        method (str): rebinning method
            "interpolation"
                Deterministic interpolation
            "listmode"
                Stochastic rebinning via conversion to listmode of energies.
                This method will internally convert input spectrum values to
                integers (if necessary) and raise a RebinWarning if the
                conversion results in a decimal precision loss.
        slopes (np.ndarray|None): (optional) an array of input bin slopes for
            quadratic interpolation (1D or 2D)
            (only applies for "interpolation" method)
            [num_spectra, num_bins_in + 1] or [num_bins_in + 1]
        zero_pad_warnings (boolean): warn when edge overlap results in
            appending empty bins

    Raises:
        AssertionError: for bad input arguments

    Returns:
        The rebinned spectrum/a
    """
    method = method.lower()
    # Cast data types and check listmode input
    if method == "listmode":
        if not np.all(in_spectra >= 0.):
            raise RebinError('Cannot rebin spectra with negative values with '
                             'listmode method')
        if np.all(in_spectra < 1):
            raise RebinError('Cannot rebin spectra with all values less than '
                             'one with listmode method')
        if np.issubdtype(in_spectra.dtype.type, np.floating):
            if np.allclose(in_spectra, np.round(in_spectra)):
                # Don't warn in the case of floats which round to integers
                pass
            else:
                warnings.warn(
                    'Argument in_spectra contains float value(s) which ' +
                    'will have decimal precision loss when converting to ' +
                    'integers for rebin method listmode.',
                    RebinWarning)
            in_spectra = np.asarray(np.round(in_spectra), dtype=np.int64)
    else:
        in_spectra = np.asarray(in_spectra, dtype=np.float64)
    in_edges = np.asarray(in_edges, np.float64)
    out_edges = np.asarray(out_edges, np.float64)
    if slopes is None:
        slopes = np.zeros_like(in_spectra, dtype=np.float64)
    else:
        slopes = np.asarray(slopes, dtype=np.float64)
    # Broadcast 1D -> 2D if necessary
    in_edges = _broadcast(in_edges, in_spectra.shape)
    slopes = _broadcast(slopes, in_spectra.shape)
    # Check dimensions
    _check_ndim(in_spectra, {1, 2}, 'in_spectra')
    _check_ndim(in_edges, {1, 2}, 'in_edges')
    _check_ndim(slopes, {1, 2}, 'slopes')
    _check_ndim(out_edges, 1, 'out_edges')
    # Check shape
    _check_shape(in_spectra, in_edges, 'in_spectra', 'in_edges', edges=True)
    _check_shape(in_spectra, slopes, 'in_spectra', 'slopes')
    # Check for increasing bin structure
    _check_monotonic_increasing(in_edges, 'in_edges')
    _check_monotonic_increasing(out_edges, 'out_edges')
    # Check for bin structure overlap
    _check_any_overlap(in_edges, out_edges)
    if zero_pad_warnings:
        _check_partial_overlap(in_edges, out_edges)
    # Specific calls to different JIT-ed rebinning methods
    if method == "interpolation":
        if in_spectra.ndim == 2:
            return _rebin2d_interpolation(
                in_spectra, in_edges, out_edges, slopes)
        elif in_spectra.ndim == 1:
            return _rebin_interpolation(
                in_spectra, in_edges, out_edges, slopes)
    elif method == "listmode":
        if in_spectra.ndim == 2:
            return _rebin2d_listmode(in_spectra, in_edges, out_edges, slopes)
        elif in_spectra.ndim == 1:
            return _rebin_listmode(in_spectra, in_edges, out_edges, slopes)
    else:
        raise ValueError("{} is not a valid rebinning method".format(method))



"""Read in a Canberra CNF file.

The CNF format specification appears to be proprietary, so this code is
based on the CNF parsing code in the xylib project:
    https://github.com/wojdyr/xylib

The relevant code is in these files:
    https://github.com/wojdyr/xylib/blob/master/xylib/util.cpp
    https://github.com/wojdyr/xylib/blob/master/xylib/canberra_cnf.cpp

"""

from __future__ import print_function
import datetime
import os
import struct
import numpy as np
from .spectrum_file import SpectrumFile, SpectrumFileParsingError


class CnfFileParsingError(SpectrumFileParsingError):
    """Failed while parsing a CNF file."""

    pass


def _from_little_endian(data, index, n_bytes):
    """Convert bytes starting from index from little endian to an integer."""
    return sum([data[index + j] << 8 * j for j in range(n_bytes)])


def _convert_date(data, index):
    """Convert 64-bit number starting at index into a date."""
    # 'data' is the number of 100ns intervals since 17 November 1858, which
    # is the start of the Modified Julian Date (MJD) epoch.
    # 3506716800 is the number of seconds between the start of the MJD epoch
    # to the start of the Unix epoch on 1 January 1970.
    d = _from_little_endian(data, index, 8)
    t = (d / 10000000.) - 3506716800
    try:
        return datetime.datetime.utcfromtimestamp(t)
    except:
        raise CnfFileParsingError('Date conversion failed')


def _convert_time(data, index):
    """Convert 64-bit number starting at index into a time."""
    # 2^(64) - number of 100ns intervals since timing started
    d = _from_little_endian(data, index, 8)
    d = (pow(2, 64) - 1) & ~d
    return d * 1.e-7


def _from_pdp11(data, index):
    """Convert 32-bit floating point in DEC PDP-11 format to a double.

    For a description of the format see:
    http://home.kpn.nl/jhm.bonten/computers/bitsandbytes/wordsizes/hidbit.htm
    """
    if (data[index + 1] & 0x80) == 0:
        sign = 1
    else:
        sign = -1
    exb = ((data[index + 1] & 0x7f) << 1) + ((data[index] & 0x80) >> 7)
    if exb == 0:
        if sign == -1:
            return np.NaN
        else:
            return 0.
    h = data[index + 2] / 256. / 256. / 256. \
        + data[index + 3] / 256. / 256. \
        + (128 + (data[index] & 0x7f)) / 256.
    return sign * h * pow(2., exb - 128.)


def _read_energy_calibration(data, index):
    """Read the four energy calibration coefficients."""
    coeff = [0., 0., 0., 0.]
    for i in range(4):
        coeff[i] = _from_pdp11(data, index + 2 * 4 + 28 + 4 * i)
    if coeff[1] == 0.:
        return None
    return coeff


class CnfFile(SpectrumFile):
    """CNF binary file parser.

    Just instantiate a class with a filename:
        spec = CnfFile(filename)

    Then the data are in
        spec.data [counts]
        spec.channels
        spec.energies
        spec.bin_edges_kev
        spec.energy_bin_widths
        spec.energy_bin_edges (deprecated)

    """

    def __init__(self, filename):
        """Initialize the CNF file."""
        super(CnfFile, self).__init__(filename)
        _, ext = os.path.splitext(self.filename)
        if ext.lower() != '.cnf':
            raise CnfFileParsingError('File extension is incorrect: ' + ext)
        # read in the data
        self.read()
        self.apply_calibration()

    def read(self, verbose=False):
        """Read in the file."""
        print('CnfFile: Reading file ' + self.filename)
        self.realtime = 0.0
        self.livetime = 0.0
        self.channels = np.array([], dtype=np.float)
        self.data = np.array([], dtype=np.float)
        self.cal_coeff = []
        data = []
        with open(self.filename, 'rb') as f:
            data_str = f.read(1)
            while data_str:
                data_int = struct.unpack('1B', data_str)
                data.append(data_int[0])
                data_str = f.read(1)
        if verbose:
            print('file size (bytes):', len(data))
        # skip first 112 bytes of file
        i = 112
        # scan for offsets
        offset_acq = 0
        offset_sam = 0
        offset_eff = 0
        offset_enc = 0
        offset_chan = 0
        for i in range(112, 128 * 1024, 48):
            offset = _from_little_endian(data, i + 10, 4)
            if (data[i + 1] == 0x20 and data[i + 2] == 0x01) \
                    or (data[i + 1] == 0) or (data[i + 2] == 0):
                if data[i] == 0:
                    if offset_acq == 0:
                        offset_acq = offset
                    else:
                        offset_enc = offset
                elif data[i] == 1:
                    if offset_sam == 0:
                        offset_sam = offset
                elif data[i] == 2:
                    if offset_eff == 0:
                        offset_eff = offset
                elif data[i] == 5:
                    if offset_chan == 0:
                        offset_chan = offset
                else:
                    pass
                if offset_acq != 0 and offset_sam != 0 \
                        and offset_eff != 0 and offset_chan != 0:
                    break
        if offset_enc == 0:
            offset_enc = offset_acq

        # extract sample information
        if (offset_sam + 48 + 80) >= len(data) or data[offset_sam] != 1 \
                or data[offset_sam + 1] != 0x20:
            if verbose:
                print(offset_sam + 48 + 80, len(data))
                print(data[offset_sam], data[offset_sam + 1])
            raise CnfFileParsingError('Sample information not found')
        else:
            sample_name = ''
            for j in range(offset_sam + 48, offset_sam + 48 + 64):
                sample_name += chr(data[j])
            if verbose:
                print('sample name: ', sample_name)
            sample_id = ''
            for j in range(offset_sam + 112, offset_sam + 112 + 64):
                sample_id += chr(data[j])
            if verbose:
                print('sample id:   ', sample_id)
            sample_type = ''
            for j in range(offset_sam + 176, offset_sam + 176 + 16):
                sample_type += chr(data[j])
            if verbose:
                print('sample type: ', sample_type)
            sample_unit = ''
            for j in range(offset_sam + 192, offset_sam + 192 + 64):
                sample_unit += chr(data[j])
            if verbose:
                print('sample unit: ', sample_unit)
            user_name = ''
            for j in range(offset_sam + 0x02d6, offset_sam + 0x02d6 + 32):
                user_name += chr(data[j])
            if verbose:
                print('user name:   ', user_name)
            sample_desc = ''
            for j in range(offset_sam + 0x036e, offset_sam + 0x036e + 256):
                sample_desc += chr(data[j])
            if verbose:
                print('sample desc: ', sample_desc)
            self.spectrum_id = sample_name
            self.sample_description = sample_desc
            self.detector_description = ''
            self.location_description = sample_id

        # extract acquisition information
        if (offset_acq + 48 + 128 + 10 + 4) >= len(data) \
                or data[offset_acq] != 0 \
                or data[offset_acq + 1] != 0x20:
            if verbose:
                print(offset_acq + 48 + 128 + 10 + 4, len(data))
                print(data[offset_acq], data[offset_acq + 1])
            raise CnfFileParsingError('Acquisition information not found')
        else:
            offset1 = _from_little_endian(data, offset_acq + 34, 2)
            offset2 = _from_little_endian(data, offset_acq + 36, 2)
            offset_pha = offset_acq + 48 + 128
            if chr(data[offset_pha + 0]) != 'P' \
                    and chr(data[offset_pha + 1]) != 'H' \
                    and chr(data[offset_pha + 2]) != 'A':
                raise CnfFileParsingError('PHA keyword not found')
            self.num_channels = 256 * \
                _from_little_endian(data, offset_pha + 10, 2)
            if self.num_channels < 256 or self.num_channels > 16384:
                raise CnfFileParsingError(
                    'Unexpected number of channels: ', self.num_channels)
            if verbose:
                print('Number of channels: ', self.num_channels)

        # extract date and time information
        offset_date = offset_acq + 48 + offset2 + 1
        if offset_date + 24 >= len(data):
            raise CnfFileParsingError('Problem with date offset')
        self.collection_start = _convert_date(data, offset_date)
        self.realtime = _convert_time(data, offset_date + 8)
        if verbose:
            print('realtime: ', self.realtime)
        self.livetime = _convert_time(data, offset_date + 16)
        if verbose:
            print('livetime: ', self.livetime)
            print('{:%Y-%m-%d %H:%M:%S}'.format(self.collection_start))
        self.collection_stop = self.collection_start \
            + datetime.timedelta(seconds=self.realtime)
        if verbose:
            print('{:%Y-%m-%d %H:%M:%S}'.format(self.collection_stop))

        # extract energy calibration information
        offset_cal = offset_enc + 48 + 32 + offset1
        if offset_enc + 48 + 32 + offset1 >= len(data):
            raise CnfFileParsingError('Problem with energy calibration offset')
        coeff = _read_energy_calibration(data, offset_enc + 48 + 32 + offset1)
        if verbose:
            print('calibration coefficients:', coeff)
        if coeff is None:
            if verbose:
                print('Energy calibration - second try')
            coeff = _read_energy_calibration(data, offset_enc + 48 + 32)
            if verbose:
                print('calibration coefficients:', coeff)
        if coeff is None:
            raise CnfFileParsingError('Energy calibration not found')
        self.cal_coeff = coeff

        # extract channel data
        if offset_chan + 512 + 4 * self.num_channels > len(data) \
                or data[offset_chan] != 5 or data[offset_chan + 1] != 0x20:
            raise CnfFileParsingError('Channel data not found')
        self.channels = np.array([], dtype=float)
        self.data = np.array([], dtype=float)
        for i in range(0, 2):
            y = _from_little_endian(data, offset_chan + 512 + 4 * i, 4)
            if y == int(self.realtime) or y == int(self.livetime):
                y = 0
            self.data = np.append(self.data, y)
            self.channels = np.append(self.channels, i)
        for i in range(2, self.num_channels):
            y = _from_little_endian(data, offset_chan + 512 + 4 * i, 4)
            self.data = np.append(self.data, y)
            self.channels = np.append(self.channels, i)

        # finished
        if self.realtime <= 0.0:
            raise CnfFileParsingError(
                'Realtime not parsed correctly: {}'.format(self.realtime))
        if self.livetime <= 0.0:
            raise CnfFileParsingError(
                'Livetime not parsed correctly: {}'.format(self.livetime))
        if self.livetime > self.realtime:
            raise CnfFileParsingError(
                'Livetime > realtime: {} > {}'.format(
                    self.livetime, self.realtime))



""""Energy calibration classes"""

from abc import ABCMeta, abstractmethod, abstractproperty
from future.builtins import dict, super, zip
from future.utils import viewitems
import numpy as np

class EnergyCalError(Exception):
    """Base class for errors in energycal.py"""

    pass


class BadInput(EnergyCalError):
    """Error related to energy cal input"""

    pass


class EnergyCalBase(object):
    """Abstract base class for energy calibration.

    A note on nomenclature: for historic reasons, 'channels' is used in
    energycal.py for generic uncalibrated x-axis values. A 'channel' is no
    longer necessarily an integer channel number (i.e., bin) from a
    multi-channel analyzer, but could for instance be a float-type fC of charge
    collected.

    Subclasses must implement:
      _ch2kev (method)
      kev2ch (method)
      valid_coeffs (property)
      _perform_fit (method)
    """

    __metaclass__ = ABCMeta

    def __init__(self):
        """Create an empty calibration instance.

        Normally you should use from_points or from_coeffs classmethods.

        Args:
          none
        """

        self._calpoints = dict()
        self._coeffs = dict()
        # initialize fit constraints?

    @classmethod
    def from_points(cls, chlist, kevlist, include_origin=False):
        """Construct EnergyCal from calibration points.

        Args:
          chlist: list/tuple/array of the channel values of calibration points
          kevlist: list/tuple/array of the corresponding energy values [keV]
          include_origin: Default is False, if set to True will add a point
                          at zero into the fit.

        Raises:
          BadInput: for bad chlist and/or kevlist.
        """

        if chlist is None or kevlist is None:
            raise BadInput('Channel list and energy list are required')

        try:
            cond = len(chlist) != len(kevlist)
        except TypeError:
            raise BadInput('Inputs must be one dimensional iterables')
        if cond:
            raise BadInput('Channels and energies must be same length')

        cal = cls()

        if include_origin:
            cal.new_calpoint(0, 0)

        for ch, kev in zip(chlist, kevlist):
            try:
                cal.new_calpoint(ch, kev)
            except (ValueError, TypeError):
                raise BadInput('Inputs must be one dimensional iterables')
        cal.update_fit()
        return cal

    @classmethod
    def from_coeffs(cls, coeffs):
        """Construct EnergyCal from equation coefficients dict.

        Args:
          coeffs: a dict with keys equal to elements in valid_coeffs,
            and values specifying the value of the coefficient
        """

        cal = cls()

        for coeff, val in viewitems(coeffs):
            cal._set_coeff(coeff, val)

        # TODO make sure all coefficients are specified

        return cal

    @property
    def channels(self):
        """The channel values of calibration points.

        Returns:
          an np.ndarray of channel values (floats)
        """

        return np.array(list(self._calpoints.values()), dtype=float)

    @property
    def energies(self):
        """The energy values of calibration points.

        Returns:
          an np.ndarray of energy values [keV]
        """

        return np.array(list(self._calpoints), dtype=float)

    @property
    def calpoints(self):
        """The calibration points, in (ch, kev) pairs.

        Returns:
          a list of 2-element tuples of (channel, energy[keV])
        """

        return list(zip(self.channels, self.energies))

    @property
    def coeffs(self):
        """The coefficients of the current calibration curve.

        Returns:
          a dict of {coeff: value}
        """

        # TODO: if there are no coeffs, error?
        return self._coeffs

    def add_calpoint(self, ch, kev):
        """Add a calibration point (ch, kev) pair. May be new or existing.

        Args:
          ch: the channel value of the calibration point
          kev: the energy value of the calibration point [keV]
        """

        self._calpoints[float(kev)] = float(ch)

    def new_calpoint(self, ch, kev):
        """Add a new calibration point. Error if energy matches existing point.

        Args:
          ch: the channel value of the calibration point
          kev: the energy value of the calibration point [keV]

        Raises:
          EnergyCalError: if energy value already exists in calibration
        """

        if kev in self._calpoints:
            raise EnergyCalError('Calibration energy already exists')
        self.add_calpoint(ch, kev)

    def rm_calpoint(self, kev):
        """Remove a calibration point, if it exists.

        Args:
          the energy value of the point to remove [keV]
        """

        if kev in self._calpoints:
            del self._calpoints[kev]
        # TODO erroring version?

    def ch2kev(self, ch):
        """Convert channel(s) to energy value(s).

        Args:
          ch: a scalar, np.array, list or tuple of channel values

        Returns:
          the energy value(s) corresponding to the channel value(s) [keV].
            a float if input is scalar. an np.array if input is iterable
        """

        if isinstance(ch, (list, tuple)):
            ch = np.array(ch)

        return self._ch2kev(ch)

    @abstractmethod
    def _ch2kev(self, ch):
        """Convert scalar OR np.array of channel(s) to energies.

        Should use numpy ufuncs so that the input dtype doesn't matter.

        Args:
          ch: an np.array, float, or int of channel values

        Returns:
          energy values, the same size/type as ch [keV]
        """

        pass

    def kev2ch(self, kev):
        """Convert energy value(s) to channel(s).

        Args:
          kev: a scalar, np.array, list or tuple of energy values [keV]

        Returns:
          the channel value(s) corresponding to the input energies.
            a float if input is scalar. an np.array if input is iterable
        """

        if isinstance(kev, (list, tuple)):
            kev = np.array(kev)

        return self._kev2ch(kev)

    @abstractmethod
    def _kev2ch(self, kev):
        """Convert energy value(s) to channel(s).

        Should use numpy ufuncs so that the input dtype doesn't matter.

        Args:
          kev: an np.array, float, or int of energy values [keV]

        Returns:
          the channel value(s) corresponding to the input energies.
            a float if input is scalar. an np.array if input is iterable
        """

        # if this is not possible, raise a NotImplementedError ?
        pass

    @abstractproperty
    def valid_coeffs(self):
        """A list of valid coefficients for the calibration curve.

        Returns:
          a tuple of strings, the names of the coefficients for this curve
        """

        pass

    def _set_coeff(self, name, val):
        """Set a coefficient for the calibration curve.

        Args:
          name: a string, the name of the coefficient to set
          val: the value to set the coefficient to

        Raises:
          EnergyCalError: if name is not in valid_coeffs
        """

        if name in self.valid_coeffs:
            self._coeffs[name] = val
        else:
            raise EnergyCalError('Invalid coefficient name: {}'.format(name))

    def update_fit(self):
        """Compute the calibration curve from the current points.

        Raises:
          EnergyCalError: if there are too few calibration points to fit
        """

        num_coeffs = len(self.valid_coeffs)
        # TODO: free coefficients, not all coefficients
        num_points = len(self._calpoints)

        if num_points == 0:
            raise EnergyCalError('No calibration points; cannot calibrate')
        elif num_points < num_coeffs:
            raise EnergyCalError('Not enough calibration points to fit curve')
        else:
            self._perform_fit()

    @abstractmethod
    def _perform_fit(self):
        """Do the actual curve fitting."""

        pass


# TODO: dummy class for testing?


class LinearEnergyCal(EnergyCalBase):
    """
    kev = b*ch + c
    """

    @classmethod
    def from_coeffs(cls, coeffs):
        """Construct LinearEnergyCal from equation coefficients dict.

        Valid coefficient names (slope, offset):
          ('b', 'c')
          ('p1', 'p0')
          ('slope', 'offset')
          ('m', 'b')

        Args:
          coeffs: a dict with keys equal to valid coeff names,
            and values specifying the value of the coefficient
        """

        new_coeffs = {}
        if 'p0' in coeffs and 'p1' in coeffs:
            new_coeffs['b'] = coeffs['p1']
            new_coeffs['c'] = coeffs['p0']
        elif 'slope' in coeffs and 'offset' in coeffs:
            new_coeffs['b'] = coeffs['slope']
            new_coeffs['c'] = coeffs['offset']
        elif 'm' in coeffs and 'b' in coeffs:
            new_coeffs['b'] = coeffs['m']
            new_coeffs['c'] = coeffs['b']
        else:
            new_coeffs = coeffs.copy()
        cal = super().from_coeffs(new_coeffs)
        return cal

    @property
    def valid_coeffs(self):
        """A list of valid coefficients for the calibration curve.

        Returns:
          a tuple of strings, the names of the coefficients for this curve
        """

        return ('b', 'c')

    @property
    def slope(self):
        """Return the slope coefficient value."""

        try:
            return self._coeffs['b']
        except KeyError:
            raise EnergyCalError(
                'Slope coefficient not yet supplied or calculated.')

    @property
    def offset(self):
        """Return the offset coefficient value."""

        try:
            return self._coeffs['c']
        except KeyError:
            raise EnergyCalError(
                'Offset coefficient not yet supplied or calculated.')

    def _ch2kev(self, ch):
        """Convert scalar OR np.array of channel(s) to energies.

        Should use numpy ufuncs so that the input dtype doesn't matter.

        Args:
          ch: an np.array, float, or int of channel values

        Returns:
          energy values, the same size/type as ch [keV]
        """

        return self.slope * ch + self.offset

    def _kev2ch(self, kev):
        """Convert energy value(s) to channel(s).

        Args:
          kev: an np.array, float, or int of energy values [keV]

        Returns:
          the channel value(s) corresponding to the input energies.
            a float if input is scalar. an np.array if input is iterable
        """

        return (kev - self.offset) / self.slope

    def _perform_fit(self):
        """Do the actual curve fitting."""

        b, c = np.polyfit(self.channels, self.energies, 1)
        self._set_coeff('b', b)
        self._set_coeff('c', c)



"""Read in an Ortec SPC file."""

from __future__ import print_function
import os
import struct
import dateutil.parser
import numpy as np
from .spectrum_file import SpectrumFile, SpectrumFileParsingError


class SpcFileParsingError(SpectrumFileParsingError):
    """Failed while parsing an SPC file."""

    pass


class SpcFile(SpectrumFile):
    """SPC binary file parser.

    Just instantiate a class with a filename:
        spec = SpcFile(filename)

    Then the data are in
        spec.data [counts]
        spec.channels
        spec.energies
        spec.bin_edges_kev
        spec.energy_bin_widths
        spec.energy_bin_edges (deprecated)

    ORTEC's SPC file format is divided into records of 128 bytes each. The
    specifications for what each record should contain can be found on pages
    29--44 of this document:
        http://www.ortec-online.com/download/ortec-software-file-structure-manual.pdf

    In the example file, not all of the records that are supposed to be in
    the files seem to be there, and so this code may depart in some places
    from the specification.

    """

    SPC_FORMAT_BEGINNING = [
        # Record 1
        [
            ['INFTYP', 'H'],
            ['FILTYP', 'H'],
            ['3', 'H'],
            ['4', 'H'],
            ['ACQIRP', 'H'],
            ['SAMDRP', 'H'],
            ['DETDRP', 'H'],
            ['EBRDESC', 'H'],
            ['ANARP1', 'H'],
            ['ANARP2', 'H'],
            ['ANARP3', 'H'],
            ['ANARP4', 'H'],
            ['SRPDES', 'H'],
            ['IEQDESC', 'H'],
            ['GEODES', 'H'],
            ['MPCDESC', 'H'],
            ['CALDES', 'H'],
            ['CALRP1', 'H'],
            ['CALRP2', 'H'],
            ['EFFPRP', 'H'],
            ['ROIRP1', 'H'],
            ['22', 'H'],
            ['23', 'H'],
            ['24', 'H'],
            ['25', 'H'],
            ['26', 'H'],
            ['PERPTR', 'H'],
            ['MAXRCS', 'H'],
            ['LSTREC', 'H'],
            ['EFFPNM', 'H'],
            ['SPCTRP', 'H'],
            ['SPCRCN', 'H'],
            ['SPCCHN', 'H'],
            ['ABSTCH', 'H'],
            ['ACQTIM', 'f'],
            ['ACQTI8', 'd'],
            ['SEQNUM', 'H'],
            ['MCANU', 'H'],
            ['SEGNUM', 'H'],
            ['MCADVT', 'H'],
            ['CHNSRT', 'H'],
            ['RLTMDT', 'f'],
            ['LVTMDT', 'f'],
            ['50', 'H'],
            ['51', 'H'],
            ['52', 'H'],
            ['53', 'H'],
            ['54', 'H'],
            ['55', 'H'],
            ['56', 'H'],
            ['57', 'H'],
            ['58', 'H'],
            ['59', 'H'],
            ['60', 'H'],
            ['61', 'H'],
            ['62', 'H'],
            ['RRSFCT', 'f'],
        ],
        # Unknown Record
        [['Unknown Record 1', '128B'], ],
        # Acquisition Information Record
        [
            ['Default spectrum file name', '16s'],
            ['Date', '12s'],
            ['Time', '10s'],
            ['Live Time', '10s'],
            ['Real Time', '10s'],
            ['59--90', 'B33x'],
            ['Start date of sample collection', '10s'],
            ['Start time of sample collection', '8s'],
            ['Stop date of sample collection', '10s'],
            ['Stop time of sample collection', '8s'],
        ],
        # Sample Description Record
        [['Sample Description', '128s'], ],
        # Detector Description Record
        [['Detector Description', '128s'], ],
        # First Analysis Parameter
        [
            ['Calibration ?', '16f'],
            ['Testing', '64s'],
        ],
        # Unknown Record
        [['Unknown Record 2', '128B'], ],
        # Calibration Description Record
        [['Calibration Description', '128s'], ],
        # Description Record 1
        [['Location Description Record 1', 'x127s'], ],
        # Description Record 2
        [['Location Description Record 2', '128s'], ],
        # Unknown Record
        [['Unknown Record 3', '128B'], ],
        # Unknown Record
        [['Unknown Record 4', '128B'], ],
        # Unknown Record
        [['Unknown Record 5', '128B'], ],
        # Unknown Record
        [['Unknown Record 6', '128B'], ],
        # Empty Record
        [['Empty Record 1', '128B'], ],
        # Empty Record
        [['Empty Record 2', '128B'], ],
        # Empty Record
        [['Empty Record 3', '128B'], ],
        # Empty Record
        [['Empty Record 4', '128B'], ],
        # Empty Record
        [['Empty Record 5', '128B'], ],
        # Hardware Parameters Record 1
        [['Hardware Parameters Record 1', '128s'], ],
        # Hardware Parameters Record 2
        [['Hardware Parameters Record 2', '128s'], ],
    ]

    SPC_FORMAT_END = [
        # Unknown Record
        [['Unknown Record 1', '128B'], ],
        # Unknown Record
        [['Unknown Record 2', '128B'], ],
        # Calibration parameters
        [
            ['Calibration parameter 0', 'f'],
            ['Calibration parameter 1', 'f'],
            ['Calibration parameter 2', 'f116x'],
        ],
    ]

    def __init__(self, filename):
        """Initialize the SPC file."""
        super(SpcFile, self).__init__(filename)
        _, ext = os.path.splitext(self.filename)
        if ext.lower() != '.spc':
            raise SpcFileParsingError('File extension is incorrect: ' + ext)
        self.metadata = {}
        # read in the data
        self.read()
        self.apply_calibration()

    def read(self, verbose=False):
        """Read in the file."""
        print('SpcFile: Reading file ' + self.filename)
        self.realtime = 0.0
        self.livetime = 0.0
        self.channels = np.array([], dtype=np.float)
        self.data = np.array([], dtype=np.float)
        self.cal_coeff = []
        with open(self.filename, 'rb') as f:
            # read the file in chunks of 128 bytes
            data_records = []
            binary_data = None
            while True:
                if binary_data is not None:
                    data_records.append(binary_data)
                try:
                    binary_data = f.read(128)
                except IOError:
                    raise SpcFileParsingError(
                        'Unable to read 128 bytes from file')
                if len(binary_data) < 128:
                    break
            if verbose:
                print(
                    'Done reading in SPC file.  Number of records: ',
                    len(data_records))
            if len(data_records) not in (279, 280):
                raise SpcFileParsingError(
                    'Number of data records incorrect: {}'.format(
                        len(data_records)))
            # read record data
            i_rec = 0
            for record_format in self.SPC_FORMAT_BEGINNING:
                if not (len(data_records) == 279 and
                        record_format[0][0] == 'Location Description Record 2'):
                    binary_data = data_records[i_rec]
                    i_rec += 1
                    fmt = '<'
                    for data_format in record_format:
                        fmt += data_format[1]
                    if verbose:
                        print('')
                        print('')
                        print('-' * 60)
                        print('')
                        print(record_format)
                        print(fmt)
                        print('')
                    data = struct.unpack(fmt, binary_data)
                    if verbose:
                        print('')
                        print(data)
                        print('')
                    for j, data_format in enumerate(record_format):
                        if isinstance(data[j], bytes):
                            self.metadata[data_format[0]] = \
                                data[j].decode('ascii')
                        else:
                            self.metadata[data_format[0]] = data[j]
                        if verbose:
                            print(
                                data_format[0],
                                ': ',
                                self.metadata[
                                    data_format[0]])
            # read spectrum records
            # These records are the spectrum data stored as INTEGER*4
            # numbers beginning with the channel number given and going
            # through the number of channels in the file. They are stored
            # as 64-word records, which gives 32 data channels per record.
            # They are stored sequentially, beginning with the record
            # pointer given.
            i_channel = 0
            for j in range(256):
                binary_data = data_records[i_rec]
                i_rec += 1
                N = struct.unpack('<32I', binary_data)
                # print(': ', N)
                for j, N_j in enumerate(N):
                    self.channels = np.append(self.channels, i_channel)
                    self.data = np.append(self.data, N_j)
                    i_channel += 1
            # read record data
            for record_format in self.SPC_FORMAT_END:
                binary_data = data_records[i_rec]
                i_rec += 1
                fmt = '<'
                for data_format in record_format:
                    fmt += data_format[1]
                if verbose:
                    print('')
                    print('')
                    print('-' * 60)
                    print('')
                    print(record_format)
                    print(fmt)
                    print('')
                data = struct.unpack(fmt, binary_data)
                if verbose:
                    print('')
                    print(data)
                    print('')
                for j, data_format in enumerate(record_format):
                    if isinstance(data[j], bytes):
                        self.metadata[data_format[0]] = data[j].decode('ascii')
                    else:
                        self.metadata[data_format[0]] = data[j]
                    if verbose:
                        print(
                            data_format[0],
                            ': ',
                            self.metadata[
                                data_format[0]])
        # finish the parsing
        self.sample_description = self.metadata['Sample Description']
        self.detector_description = self.metadata['Detector Description']
        if verbose:
            print(self.metadata['Start date of sample collection'])
        self.metadata['Start date of sample collection'] = \
            self.metadata['Start date of sample collection'][:-1]
        if verbose:
            print(self.metadata['Start date of sample collection'])
            print(self.metadata['Start time of sample collection'])
        self.collection_start = dateutil.parser.parse(
            self.metadata['Start date of sample collection'] + ' ' +
            self.metadata['Start time of sample collection'])
        if verbose:
            print(self.collection_start)
            print(self.metadata['Stop date of sample collection'])
        self.metadata['Stop date of sample collection'] = \
            self.metadata['Stop date of sample collection'][:-1]
        if verbose:
            print(self.metadata['Stop date of sample collection'])
            print(self.metadata['Stop time of sample collection'])
        self.collection_stop = dateutil.parser.parse(
            self.metadata['Stop date of sample collection'] + ' ' +
            self.metadata['Stop time of sample collection'])
        if verbose:
            print(self.collection_stop)
        self.location_description = \
            self.metadata['Location Description Record 1'][3:]
        self.location_description = self.location_description.split(
            '\x00\x00\x00')[0].replace('\x00', '\n')
        if len(data_records) > 279:
            self.location_description += \
                self.metadata['Location Description Record 2'].split(
                    '\x00\x00\x00')[0].replace('\x00', '\n')
        self.hardware_status = (
            self.metadata['Hardware Parameters Record 1'] +
            self.metadata['Hardware Parameters Record 2']).split(
                '\x00\x00\x00')[0].replace('\x00', '\n')
        self.livetime = float(self.metadata['Live Time'])
        self.realtime = float(self.metadata['Real Time'])
        if self.realtime <= 0.0:
            raise SpcFileParsingError(
                'Realtime not parsed correctly: {}'.format(self.realtime))
        if self.livetime <= 0.0:
            raise SpcFileParsingError(
                'Livetime not parsed correctly: {}'.format(self.livetime))
        if self.livetime > self.realtime:
            raise SpcFileParsingError(
                'Livetime > realtime: {} > {}'.format(
                    self.livetime, self.realtime))
        self.num_channels = len(self.channels)
        try:
            self.cal_coeff = [
                float(self.metadata['Calibration parameter 0']),
                float(self.metadata['Calibration parameter 1']),
                float(self.metadata['Calibration parameter 2']),
            ]
        except KeyError:
            raise SpcFileParsingError('Calibration parameters not found')



"""Becquerel: Tools for radiation spectral analysis."""

from . import core
from . import parsers
from . import tools
from .__metadata__ import __description__, __url__
from .__metadata__ import __version__, __license__, __copyright__

from .core.rebin import rebin, RebinError, RebinWarning
from .core.spectrum import Spectrum, SpectrumError, UncalibratedError
from .core.spectrum import SpectrumWarning
from .core.energycal import LinearEnergyCal, EnergyCalError, BadInput
from .core.utils import UncertaintiesError
from .core.plotting import SpectrumPlotter, PlottingError
from .core.peakfinder import (PeakFilter, PeakFilterError,
                              GaussianPeakFilter, PeakFinder, PeakFinderError)
from .core.autocal import AutoCalibrator, AutoCalibratorError

from .tools import nndc
from .tools.element import Element
from .tools.isotope import Isotope
from .tools.isotope_qty import IsotopeQuantity
from .tools import xcom
from .tools import materials

import warnings
warnings.simplefilter('default', DeprecationWarning)

__all__ = ['core', 'parsers', 'tools',
           'rebin', 'RebinError', 'RebinWarning',
           'Spectrum', 'SpectrumError', 'SpectrumWarning', 'SpectrumPlotter',
           'PlottingError', 'UncalibratedError', 'LinearEnergyCal',
           'EnergyCalError', 'BadInput', 'UncertaintiesError',
           'PeakFilter', 'PeakFilterError',
           'GaussianPeakFilter', 'PeakFinder', 'PeakFinderError',
           'AutoCalibrator', 'AutoCalibratorError',
           '__description__', '__url__', '__version__', '__license__',
           '__copyright__']



"""Query material data NIST X-Ray Mass Attenuation Coefficients website.

References:
  https://www.nist.gov/pml/x-ray-mass-attenuation-coefficients
  http://physics.nist.gov/PhysRefData/XrayMassCoef/tab1.html
  http://physics.nist.gov/PhysRefData/XrayMassCoef/tab2.html

"""

from __future__ import print_function
import requests
import pandas as pd
from .element import element_symbol
from ..core.utils import isstring, Iterable

MAX_Z = 92
N_COMPOUNDS = 48


_URL_TABLE1 = 'http://physics.nist.gov/PhysRefData/XrayMassCoef/tab1.html'
_URL_TABLE2 = 'http://physics.nist.gov/PhysRefData/XrayMassCoef/tab2.html'


class NISTMaterialsError(Exception):
    """General error for NIST materials data."""

    pass


class NISTMaterialsRequestError(NISTMaterialsError):
    """Error related to communicating with NIST or parsing the result."""

    pass


def _get_request(url):
    """Perform GET request.

    Args:
      url: URL (string) to get.

    Returns:
      requests object.

    Raises:
      NISTMaterialsRequestError: if there was a problem making the request.

    """

    req = requests.get(url)
    if not req.ok or req.reason != 'OK' or req.status_code != 200:
        raise NISTMaterialsRequestError(
            'NIST materials request failed: reason={}, status_code={}'.format(
                req.reason, req.status_code))
    return req


def fetch_element_data():
    """Retrieve data for the elements.

    Data are found in Table 1:
      http://physics.nist.gov/PhysRefData/XrayMassCoef/tab1.html

    Returns:
      A pandas DataFrame of the material data. Colums are 'Z', 'Symbol',
      'Element', 'Z_over_A', 'I_eV', and 'Density'.

    Raises:
      NISTMaterialsRequestError: if there was a problem obtaining the data.

    """
    req = _get_request(_URL_TABLE1)
    text = req.text
    # rename first two header columns
    text = text.replace(
        '<TH scope="col" COLSPAN="2"><I>Z</I></TH>',
        '<TH scope="col">Z</TH><TH scope="col">Symbol</TH>')
    # remove row that makes the table too wide
    text = text.replace('<TD COLSPAN="10"><HR SIZE="1" NOSHADE></TD>', '')
    # remove extra header columns
    text = text.replace('<TD COLSPAN="2">', '<TD>')
    text = text.replace('<TD COLSPAN="4">', '<TD>')
    text = text.replace('TD>&nbsp;</TD>', '')
    # remove extra columns in Hydrogen row
    text = text.replace('<TD ROWSPAN="92">&nbsp;</TD>', '')
    # remove open <TR> at the end of the table
    text = text.replace('</TD></TR><TR>', '</TD></TR>')
    # read HTML table into pandas DataFrame
    tables = pd.read_html(text, header=0, skiprows=[1, 2])
    if len(tables) != 1:
        raise NISTMaterialsRequestError(
            '1 HTML table expected, but found {}'.format(len(tables)))
    df = tables[0]
    if len(df) != MAX_Z:
        raise NISTMaterialsRequestError(
            '{} elements expected, but found {}'.format(MAX_Z, len(df)))
    if len(df.columns) != 6:
        raise NISTMaterialsRequestError(
            '10 columns expected, but found {} ({})'.format(
                len(df.columns), df.columns))
    # rename columns
    df.columns = ['Z', 'Symbol', 'Element', 'Z_over_A', 'I_eV', 'Density']
    return df


def convert_composition(comp):
    """Convert composition by Z into composition by symbol.

    Args:
      comp: a list of strings from the last column of Table 2, e.g.,
        ["1: 0.111898", "8: 0.888102"]

    Returns:
      A list of strings of composition by symbol, e.g.,
        ["H 0.111898", "O 0.888102"]

    Raises:
      NISTMaterialsRequestError: if there was a problem making the conversion.

    """
    comp_sym = []
    if not isinstance(comp, Iterable):
        raise NISTMaterialsRequestError(
            'Compound must be an iterable of strings: {}'.format(comp))
    for line in comp:
        if not isstring(line):
            raise NISTMaterialsRequestError(
                'Line must be a string type: {} {}'.format(line, type(line)))
        try:
            z, weight = line.split(':')
        except ValueError:
            raise NISTMaterialsRequestError(
                'Unable to split compound line: {}'.format(line))
        try:
            z = int(z)
        except ValueError:
            raise NISTMaterialsRequestError(
                'Unable to convert Z {} to integer: {}'.format(z, line))
        if z < 1 or z > MAX_Z:
            raise NISTMaterialsRequestError(
                'Z {} out of range [1, {}]: {}'.format(z, line, MAX_Z))
        comp_sym.append(element_symbol(z) + ' ' + weight.strip())
    return comp_sym


def fetch_compound_data():
    """Retrieve data for the compounds.

    Data are found in Table 2:
      http://physics.nist.gov/PhysRefData/XrayMassCoef/tab2.html

    Returns:
      A pandas DataFrame of the material data. Columns are 'Material',
      'Z_over_A', 'I_eV', 'Density', 'Composition_Z', and 'Composition_symbol'.

    Raises:
      NISTMaterialsRequestError: if there was a problem obtaining the data.

    """
    req = _get_request(_URL_TABLE2)
    text = req.text
    # remove extra header columns
    text = text.replace('<TD ROWSPAN="2">&nbsp;</TD>', '')
    # remove extra rows in header row
    text = text.replace(' ROWSPAN="2"', '')
    # remove extra columns in third header row
    text = text.replace('<TD COLSPAN="9"><HR SIZE="1" NOSHADE></TD>', '')
    # remove extra columns in the first material row
    text = text.replace('<TD ROWSPAN="50"> &nbsp; </TD>', '')
    # replace <BR> symbols in composition lists with semicolons
    text = text.replace('<BR>', ';')
    # read HTML table into pandas DataFrame
    tables = pd.read_html(text, header=0, skiprows=[1, 2])
    if len(tables) != 1:
        raise NISTMaterialsRequestError(
            '1 HTML table expected, but found {}'.format(len(tables)))
    df = tables[0]
    if len(df) != N_COMPOUNDS:
        raise NISTMaterialsRequestError(
            '{} compounds expected, but found {}'.format(N_COMPOUNDS, len(df)))
    if len(df.columns) != 5:
        raise NISTMaterialsRequestError(
            '5 columns expected, but found {} ({})'.format(
                len(df.columns), df.columns))
    # rename columns
    df.columns = ['Material', 'Z_over_A', 'I_eV', 'Density', 'Composition_Z']
    # clean up Z composition
    df['Composition_Z'] = [
        [line.strip() for line in comp.split(';')]
        for comp in df['Composition_Z']]
    # create a column of compositions by symbol (for use with fetch_xcom_data)
    df['Composition_symbol'] = [
        convert_composition(comp) for comp in df['Composition_Z']]
    return df



#!/usr/bin/env python
"""Becquerel: Tools for radiation spectral analysis."""

from __future__ import print_function
from setuptools import setup, find_packages

NAME = 'becquerel'

MAJOR = 0
MINOR = 2
MICRO = 4
VERSION = '{}.{}.{}'.format(MAJOR, MINOR, MICRO)

DESCRIPTION = __doc__.split('\n')[0].split(': ')[-1]
URL = 'https://github.com/lbl-anp/becquerel'
MAINTAINER = "The Becquerel Development Team"
EMAIL = "becquerel-dev@lbl.gov"

# classifiers from list at https://pypi.org/classifiers/
CLASSIFIERS = """\
Development Status :: 3 - Alpha
Intended Audience :: Science/Research
License :: Other/Proprietary License
Operating System :: OS Independent
Programming Language :: Python
Programming Language :: Python :: 2
Programming Language :: Python :: 3
Topic :: Scientific/Engineering
Topic :: Scientific/Engineering :: Physics
"""

with open('README.md', 'r') as fh:
    README = fh.read()
# remove package title from description
README = '\n'.join(README.split('\n')[2:])

with open('CONTRIBUTING.md', 'r') as fh:
    CONTRIBUTING = fh.read()

with open('LICENSE.txt', 'r') as fh:
    LICENSE = fh.read()

with open('requirements.txt', 'r') as fh:
    REQUIREMENTS = fh.read()

# make long description from README and CONTRIBUTING
# but move copyright notice to the end
LONG_DESCRIPTION, COPYRIGHT = README.split('## Copyright Notice')
LONG_DESCRIPTION += '\n' + CONTRIBUTING
LONG_DESCRIPTION += '\n' + '## Copyright Notice' + COPYRIGHT

# write metadata to a file that will be imported by becquerel
with open('becquerel/__metadata__.py', 'w') as f:
    print('"""Becquerel package metadata."""', file=f)
    print('', file=f)
    print('__description__ = "{}"'.format(DESCRIPTION), file=f)
    print('__url__ = "{}"'.format(URL), file=f)
    print('__version__ = "{}"'.format(VERSION), file=f)
    print('__license__ = """{}"""'.format(LICENSE), file=f)
    print('__copyright__ = """{}"""'.format(COPYRIGHT), file=f)

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url=URL,
    download_url=URL + '/releases',
    maintainer=MAINTAINER,
    maintainer_email=EMAIL,
    classifiers=[_f for _f in CLASSIFIERS.split('\n') if _f],
    platforms='any',
    packages=find_packages(),
    python_requires='>=2.6',
    install_requires=[_f for _f in REQUIREMENTS.split('\n') if _f],
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'pytest-cov'],
    license="Other/Proprietary License (see LICENSE.txt)",
)



# -*- coding: utf-8 -*-

"""Spectral peak search using convolutions."""

from __future__ import print_function
import matplotlib.pyplot as plt
import numpy as np
import warnings
from .spectrum import Spectrum


class PeakFilterError(Exception):
    """Base class for errors in PeakFilter."""

    pass


class PeakFilter(object):
    """An energy-dependent kernel that can be convolved with a spectrum.

    To detect lines, a kernel should have a positive component in the center
    and negative wings to subtract the continuum, e.g., a Gaussian or a boxcar:

    +2|     ┌───┐     |
      |     │   │     |
     0|─┬───┼───┼───┬─|
    -1| └───┘   └───┘ |

    The positive part is meant to detect the peak, and the negative part to
    sample the continuum under the peak.

    The kernel should sum to 0.

    The width of the kernel scales proportionally to the square root
    of the x values (which could be energy, ADC channels, fC of charge
    collected, etc.), with a minimum value set by 'fwhm_at_0'.

    """

    def __init__(self, ref_x, ref_fwhm, fwhm_at_0=1.):
        """Initialize with a reference line position and FWHM in x-values."""
        if ref_x <= 0:
            raise PeakFilterError('Reference x must be positive')
        if ref_fwhm <= 0:
            raise PeakFilterError('Reference FWHM must be positive')
        if fwhm_at_0 < 0:
            raise PeakFilterError('FWHM at 0 must be non-negative')
        self.ref_x = float(ref_x)
        self.ref_fwhm = float(ref_fwhm)
        self.fwhm_at_0 = float(fwhm_at_0)

    def fwhm(self, x):
        """Calculate the expected FWHM at the given x value."""
        # f(x)^2 = f0^2 + k x^2
        # f1^2 = f0^2 + k x1^2
        # k = (f1^2 - f0^2) / x1^2
        # f(x)^2 = f0^2 + (f1^2 - f0^2) (x/x1)^2
        f0 = self.fwhm_at_0
        f1 = self.ref_fwhm
        x1 = self.ref_x
        fwhm_sqr = f0**2 + (f1**2 - f0**2) * (x / x1)**2
        return np.sqrt(fwhm_sqr)

    def kernel(self, x, edges):
        """Generate the kernel for the given x value."""
        raise NotImplementedError

    def kernel_matrix(self, edges):
        """Build a matrix of the kernel evaluated at each x value."""
        n_channels = len(edges) - 1
        kern = np.zeros((n_channels, n_channels))
        for i, x in enumerate(edges[:-1]):
            kern[:, i] = self.kernel(x, edges)
        kern_pos = +1 * kern.clip(0, np.inf)
        kern_neg = -1 * kern.clip(-np.inf, 0)
        # normalize negative part to be equal to the positive part
        kern_neg *= kern_pos.sum(axis=0) / kern_neg.sum(axis=0)
        return kern_pos - kern_neg

    def plot_matrix(self, edges):
        """Plot the matrix of kernels evaluated across the x values."""
        n_channels = len(edges) - 1
        kern_mat = self.kernel_matrix(edges)
        kern_min = kern_mat.min()
        kern_max = kern_mat.max()
        kern_min = min(kern_min, -1 * kern_max)
        kern_max = max(kern_max, -1 * kern_min)

        plt.imshow(
            kern_mat.T[::-1, :], cmap=plt.get_cmap('bwr'),
            vmin=kern_min, vmax=kern_max,
            extent=[n_channels, 0, 0, n_channels])
        plt.colorbar()
        plt.xlabel('Input x')
        plt.ylabel('Output x')
        plt.gca().set_aspect('equal')

    def convolve(self, edges, data):
        """Convolve this kernel with the data."""
        kern_mat = self.kernel_matrix(edges)
        kern_mat_pos = +1 * kern_mat.clip(0, np.inf)
        kern_mat_neg = -1 * kern_mat.clip(-np.inf, 0)
        peak_plus_bkg = np.dot(kern_mat_pos, data)
        bkg = np.dot(kern_mat_neg, data)
        signal = np.dot(kern_mat, data)
        noise = np.sqrt(np.dot(kern_mat**2, data))
        snr = np.zeros_like(signal)
        snr[noise > 0] = signal[noise > 0] / noise[noise > 0]
        return peak_plus_bkg, bkg, signal, noise, snr


def _gaussian0(x, mean, sigma):
    """Gaussian function."""
    z = (x - mean) / sigma
    return np.exp(-z**2 / 2.)


def _gaussian1(x, mean, sigma):
    """First derivative of a gaussian."""
    z = (x - mean)
    return -1 * z * _gaussian0(x, mean, sigma)


class GaussianPeakFilter(PeakFilter):
    """A spectral kernel that is the second derivative of a Gaussian.

    g0(x, u, s) = 1 / (sqrt(2 pi) s) * exp(-(x-u)^2/(2 s^2))
    g1(x, u, s) = -(x - u) / (s^2) * g0
    g2(x, u, s) = (x - u)^2 / (s^4) * g0 - (1/s^2) * g0
                = g0(x,u,s) ((x-u)^2/s^2 - 1) / s^2

    integral(-g2, x=x0..x1)
        = (-g1) for x=x0..x1
        = g1(x0) - g1(x1)

    """

    def kernel(self, x, edges):
        """Generate the kernel for the given x value."""
        fwhm = self.fwhm(x)
        sigma = fwhm / 2.355
        g1_x0 = _gaussian1(edges[:-1], x, sigma)
        g1_x1 = _gaussian1(edges[1:], x, sigma)
        kernel = g1_x0 - g1_x1
        return kernel


class PeakFinderError(Exception):
    """Base class for errors in PeakFinder."""

    pass


class PeakFinder(object):
    """Find peaks in a spectrum after convolving it with a kernel."""

    def __init__(self, spectrum, kernel, min_sep=5, fwhm_tol=(0.5, 1.5)):
        """Initialize with a spectrum and kernel."""
        if min_sep <= 0:
            raise PeakFinderError(
                'Minimum x separation must be positive')
        self.min_sep = min_sep
        self.fwhm_tol = tuple(fwhm_tol)
        self.spectrum = None
        self.kernel = None
        self.snr = []
        self._peak_plus_bkg = []
        self._bkg = []
        self._signal = []
        self._noise = []
        self.centroids = []
        self.snrs = []
        self.fwhms = []
        self.integrals = []
        self.backgrounds = []
        self.calculate(spectrum, kernel)

    @property
    def channels(self):
        warnings.warn('channels is deprecated and will be removed in a future '
                      'release. Use centroids instead.', DeprecationWarning)
        return self.centroids

    def reset(self):
        """Restore peak finder to pristine starting condition."""
        self.centroids = []
        self.snrs = []
        self.fwhms = []
        self.integrals = []
        self.backgrounds = []

    def sort_by(self, arr):
        """Sort peaks by the provided array."""
        if len(arr) != len(self.centroids):
            raise PeakFinderError(
                'Sorting array has length {} but must have length {}'.format(
                    len(arr), len(self.centroids)))
        self.centroids = np.array(self.centroids)
        self.snrs = np.array(self.snrs)
        self.fwhms = np.array(self.fwhms)
        self.integrals = np.array(self.integrals)
        self.backgrounds = np.array(self.backgrounds)
        i = np.argsort(arr)
        self.centroids = list(self.centroids[i])
        self.snrs = list(self.snrs[i])
        self.fwhms = list(self.fwhms[i])
        self.integrals = list(self.integrals[i])
        self.backgrounds = list(self.backgrounds[i])

    def calculate(self, spectrum, kernel):
        """Calculate the convolution of the spectrum with the kernel."""
        if not isinstance(spectrum, Spectrum):
            raise PeakFinderError(
                'Argument must be a Spectrum, not {}'.format(type(spectrum)))
        if not isinstance(kernel, PeakFilter):
            raise PeakFinderError(
                'Argument must be a PeakFilter, not {}'.format(type(kernel)))
        self.spectrum = spectrum
        self.kernel = kernel
        self.snr = np.zeros(len(self.spectrum))

        bin_edges = self.spectrum.bin_edges_raw

        # calculate the convolution
        peak_plus_bkg, bkg, signal, noise, snr = \
            self.kernel.convolve(bin_edges, self.spectrum.counts_vals)
        self._peak_plus_bkg = peak_plus_bkg
        self._bkg = bkg
        self._signal = signal
        self._noise = noise
        self.snr = snr.clip(0)
        self.reset()

    def add_peak(self, xpeak):
        """Add a peak at xpeak to list if it is not already there."""
        bin_edges = self.spectrum.bin_edges_raw

        xmin = bin_edges.min()
        xmax = bin_edges.max()

        if xpeak < xmin or xpeak > xmax:
            raise PeakFinderError(
                'Peak x {} is outside of range {}-{}'.format(
                    xpeak, xmin, xmax))
        is_new_x = True
        for cent in self.centroids:
            if abs(xpeak - cent) <= self.min_sep:
                is_new_x = False
        if is_new_x:
            # estimate FWHM using the second derivative
            # snr(x) = snr(x0) - 0.5 d2snr/dx2(x0) (x-x0)^2
            # 0.5 = 1 - 0.5 d2snr/dx2 (fwhm/2)^2 / snr0
            # 1 = d2snr/dx2 (fwhm/2)^2 / snr0
            # fwhm = 2 sqrt(snr0 / d2snr/dx2)
            xbin = self.spectrum.find_bin_index(xpeak, use_kev=False)
            fwhm0 = self.kernel.fwhm(xpeak)
            bw = self.spectrum.bin_widths_raw[0]
            h = int(max(1, 0.2 * fwhm0 / bw))
            d2 = (1 * self.snr[xbin - h]
                  - 2 * self.snr[xbin]
                  + 1 * self.snr[xbin + h]) / h**2 / bw**2
            if d2 >= 0:
                raise PeakFinderError(
                    'Second derivative must be negative at peak')
            d2 *= -1
            fwhm = 2 * np.sqrt(self.snr[xbin] / d2)
            self.fwhms.append(fwhm)
            # add the peak if it has a similar FWHM to the kernel's FWHM
            if self.fwhm_tol[0] * fwhm0 <= fwhm <= self.fwhm_tol[1] * fwhm0:
                self.centroids.append(xpeak)
                self.snrs.append(self.snr[xbin])
                self.fwhms.append(fwhm)
                self.integrals.append(self._signal[xbin])
                self.backgrounds.append(self._bkg[xbin])
        # sort the peaks by centroid
        self.sort_by(self.centroids)

    def plot(self, facecolor='red', linecolor='red', alpha=0.5, peaks=True):
        """Plot the peak signal-to-noise ratios calculated using the kernel."""
        bin_edges = self.spectrum.bin_edges_raw

        if facecolor is not None:
            plt.fill_between(
                bin_edges[:-1], self.snr, 0,
                color=facecolor, alpha=alpha)
        if linecolor is not None:
            plt.plot(bin_edges[:-1], self.snr, '-', color=linecolor)
        if peaks:
            for cent, snr, fwhm in zip(self.centroids, self.snrs, self.fwhms):
                plt.plot([cent] * 2, [0, snr], 'b-', lw=1.5)
                plt.plot(cent, snr, 'bo')
                plt.plot(
                    [cent - fwhm / 2, cent + fwhm / 2], [snr / 2] * 2,
                    'b-', lw=1.5)
        plt.xlim(0, bin_edges.max())
        plt.ylim(0)
        plt.xlabel('x')
        plt.ylabel('SNR')

    def find_peak(self, xpeak, frac_range=(0.8, 1.2), min_snr=2):
        """Find the highest SNR peak within f0*xpeak and f1*xpeak."""
        bin_edges = self.spectrum.bin_edges_raw
        bin_centers = self.spectrum.bin_centers_raw
        xmin = bin_edges[0]
        xmax = bin_edges[-1]

        if xpeak < xmin or xpeak > xmax:
            raise PeakFinderError(
                'Guess xpeak {} is outside of range {}-{}'.format(
                    xpeak, xmin, xmax))
        if frac_range[0] < 0 or frac_range[0] > 1 or frac_range[1] < 1 or \
                frac_range[0] > frac_range[1]:
            raise PeakFinderError(
                'Fractional range {}-{} is invalid'.format(*frac_range))
        if min_snr < 0:
            raise PeakFinderError(
                'Minimum SNR {:.3f} must be > 0'.format(min_snr))
        if self.snr.max() < min_snr:
            raise PeakFinderError(
                'SNR threshold is {:.3f} but maximum SNR is {:.3f}'.format(
                    min_snr, self.snr.max()))
        x0 = frac_range[0] * xpeak
        x1 = frac_range[1] * xpeak
        x_range = (x0 <= bin_edges[:-1]) & (bin_edges[:-1] <= x1)
        peak_snr = self.snr[x_range].max()
        if peak_snr < min_snr:
            raise PeakFinderError(
                'No peak found in range {}-{} with SNR > {}'.format(
                    x0, x1, min_snr))

        peak_index = np.where((self.snr == peak_snr) & x_range)[0][0]
        peak_x = bin_centers[peak_index]
        self.add_peak(peak_x)
        return peak_x

    def find_peaks(self, xmin=None, xmax=None, min_snr=2, max_num=40):
        """Find the highest SNR peaks in the data."""
        bin_edges = self.spectrum.bin_edges_raw
        bin_centers = self.spectrum.bin_centers_raw

        if xmin is None:
            xmin = bin_edges.min()
        if xmax is None:
            xmax = bin_edges.max()
        if xmin < bin_edges.min() or \
                xmin > bin_edges.max() or \
                xmax > bin_edges.max() or \
                xmax < bin_edges.min() or \
                xmin > xmax:
            raise PeakFinderError(
                'x-axis range {}-{} is invalid'.format(xmin, xmax))
        if min_snr < 0:
            raise PeakFinderError(
                'Minimum SNR {:.3f} must be > 0'.format(min_snr))
        if self.snr.max() < min_snr:
            raise PeakFinderError(
                'SNR threshold is {:.3f} but maximum SNR is {:.3f}'.format(
                    min_snr, self.snr.max()))
        max_num = int(max_num)
        if max_num < 1:
            raise PeakFinderError(
                'Must keep at least 1 peak, not {}'.format(max_num))
        # calculate the first derivative and second derivatives of the SNR
        d1 = (self.snr[2:] - self.snr[:-2]) / 2
        d1 = np.append(0, d1)
        d1 = np.append(d1, 0)
        d2 = self.snr[2:] - 2 * self.snr[1:-1] + self.snr[:-2]
        d2 = np.append(0, d2)
        d2 = np.append(d2, 0)
        # find maxima
        peak = (d1[2:] < 0) & (d1[:-2] > 0) & (d2[1:-1] < 0)
        peak = np.append(False, peak)
        peak = np.append(peak, False)
        # select peaks using SNR and centroid criteria
        peak &= (min_snr <= self.snr)
        peak &= (xmin <= bin_edges[:-1])
        peak &= (bin_edges[:-1] <= xmax)
        for x in bin_centers[peak]:
            self.add_peak(x)
        # reduce number of centroids to a maximum number max_n of highest SNR
        self.sort_by(np.array(self.snrs))
        self.centroids = self.centroids[-max_num:]
        self.snrs = self.snrs[-max_num:]
        self.fwhms = self.fwhms[-max_num:]
        self.integrals = self.integrals[-max_num:]
        self.backgrounds = self.backgrounds[-max_num:]
        # sort by centroid
        self.sort_by(self.centroids)



"""Query half-life and decay data from the National Nuclear Data Center.

References:
  http://www.nndc.bnl.gov
  http://www.nndc.bnl.gov/nudat2/indx_sigma.jsp
  http://www.nndc.bnl.gov/nudat2/indx_dec.jsp

"""

from __future__ import print_function
from future.builtins import super
import warnings
import numpy as np
import requests
import pandas as pd
import uncertainties
from ..core.utils import isstring, ResourceWarning


PARITIES = ['+', '-', 'any']


DECAYRAD_DECAY_MODE = {
    'any': 'ANY',
    'internal transition': 'IT',
    'it': 'IT',
    'beta-': 'B-',
    'b-': 'B-',
    'electron capture beta+': 'ECBP',
    'ecbp': 'ECBP',
    'ecb+': 'ECBP',
    'ec+b+': 'ECBP',
    'electron capture': 'ECBP',
    'ec': 'ECBP',
    'beta+': 'ECBP',
    'b+': 'ECBP',
    'neutron': 'N',
    'n': 'N',
    'proton': 'P',
    'p': 'P',
    'alpha': 'A',
    'a': 'A',
    'spontaneous fission': 'SF',
    'sf': 'SF',
}


WALLET_DECAY_MODE = dict(DECAYRAD_DECAY_MODE)
WALLET_DECAY_MODE.update({
    'double beta': 'DB',
    'bb': 'DB',
    'cluster': 'C',
    'c': 'C',
    'beta-delayed neutron': 'DN',
    'b-delayed n': 'DN',
    'bdn': 'DN',
    'beta-delayed proton': 'DP',
    'b-delayed p': 'DP',
    'bdp': 'DP',
    'beta-delayed alpha': 'DA',
    'b-delayed a': 'DA',
    'bda': 'DA',
    'beta-delayed fission': 'DF',
    'b-delayed f': 'DF',
    'bdf': 'DF',
})


DECAYRAD_RADIATION_TYPE = {
    'any': 'ANY',
    'gamma': 'G',
    'g': 'G',
    'beta-': 'BM',
    'b-': 'BM',
    'beta+': 'BP',
    'b+': 'BP',
    'electron': 'E',
    'e': 'E',
    'proton': 'P',
    'p': 'P',
    'alpha': 'A',
    'a': 'A',
}


class NNDCError(Exception):
    """General NNDC request error."""

    pass


class NoDataFound(NNDCError):
    """No datasets were found within the specified search."""


class NNDCInputError(NNDCError):
    """Error related to the user input to an NNDC query."""

    pass


class NNDCRequestError(NNDCError):
    """Error related to communicating with NNDC or parsing the result."""

    pass


def _parse_headers(headers):
    """Parse table headers and ensure they are unique.

    Args:
      headers: a list of column header strings.

    Returns:
      a new list of strings where abbreviations have been expanded.

    Raises:
      NNDCRequestError: if there was a problem parsing the headers.

    """

    headers_new = []
    # reformat column headers if needed
    for j, hd in enumerate(headers):
        # rename so always have T1/2 (s)
        if hd == 'T1/2 (num)' or hd == 'T1/2 (seconds)':
            hd = 'T1/2 (s)'
        # for uncertainties, add previous column header to it
        if j > 0 and 'Unc' in hd:
            hd = headers[j - 1] + ' ' + hd
        if 'Unc' in hd and 'Unc.' not in hd:
            hd = hd.replace('Unc', 'Unc.')
        # expand abbreviated headers
        if 'Energy' in hd and 'Energy Level' not in hd:
            hd = hd.replace('Energy', 'Energy Level')
        if 'Par. Elevel' in hd:
            hd = hd.replace('Par. Elevel', 'Parent Energy Level')
        if 'Abund.' in hd:
            hd = hd.replace('Abund.', 'Abundance (%)')
        if 'Ene.' in hd:
            hd = hd.replace('Ene.', 'Energy')
        if 'Int.' in hd:
            hd = hd.replace('Int.', 'Intensity (%)')
        if 'Dec' in hd and 'Decay' not in hd:
            hd = hd.replace('Dec', 'Decay')
        if 'Rad' in hd and 'Radiation' not in hd:
            hd = hd.replace('Rad', 'Radiation')
        if 'EP' in hd:
            hd = hd.replace('EP', 'Endpoint')
        if 'Mass Exc' in hd and 'Mass Excess' not in hd:
            hd = hd.replace('Mass Exc', 'Mass Excess')
        headers_new.append(hd)
    if len(set(headers_new)) != len(headers_new):
        raise NNDCRequestError(
            'Duplicate headers after parsing\n' +
            '    Original headers: "{}"\n'.format(headers) +
            '    Parsed headers:   "{}"'.format(headers_new))
    return headers_new


def _parse_table(text):
    """Parse table contained in the text into a dictionary.

    Args:
      text: a string containing an HTML table from the NNDC request

    Returns:
      a dictionary of lists keyed by the column headers.

    Raises:
      NNDCRequestError: if unable to parse the table.

    """

    text = str(text)
    try:
        text = text.split('<pre>')[1]
        text = text.split('</pre>')[0]
        text = text.split('To save this output')[0]
        lines = text.split('\n')
    except Exception as exc:
        raise NNDCRequestError('Unable to parse text:\n{}\n{}'.format(exc,
                                                                      text))
    table = {}
    headers = None
    for line in lines:
        tokens = line.split('\t')
        tokens = [t.strip() for t in tokens]
        if len(tokens) <= 1:
            continue
        if headers is None:
            headers = tokens
            headers = _parse_headers(headers)
            for header in headers:
                table[header] = []
        else:
            if len(tokens) != len(headers):
                raise NNDCRequestError(
                    'Too few data in table row\n' +
                    '    Headers: "{}"\n'.format(headers) +
                    '    Row:     "{}"'.format(tokens))
            for header, token in zip(headers, tokens):
                table[header].append(token)
    return table


def _parse_float_uncertainty(x, dx):
    """Parse a string and its uncertainty into a float or ufloat.

    Examples:
      >>> _parse_float_uncertainty('257.123', '0.005')
      257.123+/-0.005
      >>> _parse_float_uncertainty('8', '')
      8.0

    Args:
      x: a string representing the nominal value of the quantity.
      dx: a string representing the uncertainty of the quantity.

    Returns:
      a float (if dx == '') or a ufloat.

    Raises:
      NNDCRequestError: if values cannot be parsed.

    """

    if not isstring(x):
        raise NNDCRequestError('Value must be a string: {}'.format(x))
    if not isstring(dx):
        raise NNDCRequestError('Uncertainty must be a string: {}'.format(dx))
    # ignore percents
    if '%' in x:
        x = x.replace('%', '')
    # ignore unknown ground state levels (X, Y, Z, W)
    for sym in ['X', 'Y', 'Z', 'W']:
        if '+' + sym in x:
            x = x.replace('+' + sym, '')
        elif x == sym:
            x = '0'
    # handle special ENSDF abbreviations, e.g.,
    # http://www.iaea.org/inis/collection/NCLCollectionStore/_Public/14/785/14785563.pdf
    # "One of the following expressions:
    #   LT, GT, LE, GE, AP, CA, SY
    # for less than, greater than, less than or equal to greater
    # than or equal to. approximately equal to, calculated, and
    # from systematics, respectively."
    for sym in ['*', '<', '>', '=', '~', '?', '@', '&', 'P', 'N']:
        while sym in x:
            x = x.replace(sym, '')
    # correct specific typos in the database
    if 'E-11 0' in x:
        x = x.replace('E-11 0', 'E-11')
    if 'E-12 0' in x:
        x = x.replace('E-12 0', 'E-12')
    if '0.0000 1' in x:
        x = x.replace('0.0000 1', '0.0000')
    if '2 .8E-7' in x:
        x = x.replace('2 .8E-7', '2.8E-7')
    if '8 .0E-E5' in x:
        x = x.replace('8 .0E-E5', '8.0E-5')
    # handle blank or missing data
    if x == '' or x == ' ':
        return None
    if '****' in dx:
        dx = ''
    elif dx in ['LT', 'GT', 'LE', 'GE', 'AP', 'CA', 'SY']:
        dx = ''
    try:
        x2 = float(x)
    except ValueError:
        raise NNDCRequestError(
            'Value cannot be parsed as float: "{}"'.format(x))
    if dx == '':
        return x2
    # handle multiple exponents with some uncertainties, e.g., "7E-4E-5"
    tokens = dx.split('E')
    if len(tokens) == 3:
        dx = 'E'.join(tokens[:2])
        factor = pow(10., int(tokens[2]))
    else:
        factor = 1.
    try:
        dx2 = float(dx) * factor
    except ValueError:
        raise NNDCRequestError(
            'Uncertainty cannot be parsed as float: "{}"'.format(dx))
    return uncertainties.ufloat(x2, dx2)


def _format_range(x_range):
    """Return two strings for the two range elements, blank if not finite.

    Args:
      x_range: an iterable of 2 range limits, which can be numbers
        or inf/NaN/None.

    Returns:
      an iterable of 2 strings.

    Raises:
      NNDCInputError: if x_range is not an iterable of length 2.

    """

    try:
        x1, x2 = x_range
    except (TypeError, ValueError):
        raise NNDCInputError(
            'Range keyword arg must have two elements: "{}"'.format(x_range))
    try:
        if np.isfinite(x1):
            x1 = '{}'.format(x1)
        else:
            x1 = ''
    except TypeError:
        x1 = ''
    try:
        if np.isfinite(x2):
            x2 = '{}'.format(x2)
        else:
            x2 = ''
    except TypeError:
        x2 = ''
    return x1, x2


class _NNDCQuery(object):
    """National Nuclear Data Center database query base class.

    Args:
      perform: a boolean dictating whether to immediately perform the query.
      nuc     : (str) : the name of the isotope (e.g., 'Co-60')
      z, a, n : (int) : Z, A, N of the isotope
      z_range, etc. : (tuple of int) : range of Z, A, or N
      z_any, etc. : (bool) : whether any Z, A, or N is considered
      z_odd, etc. : (bool) : only odd Z, A, or N
      z_even, etc.: (bool) : only even Z, A, or N
      t_range : (tuple of float) : range of isotope half-lives in seconds

    Raises:
      NNDCInputError: if there is a problem with the input.
      NNDCRequestError: if there was a problem with the data requested.

    """

    _URL = ''
    _DATA = {
        'spnuc': '',           # specify parent ('name', 'zan', or 'zanrange')
        'nuc': '',             # isotope name (use with 'name')
        'z': '',               # Z or element (use with 'zan')
        'zmin': '',            # Z min        (use with 'zanrange')
        'zmax': '',            # Z max        (use with 'zanrange')
        'a': '',               # A            (use with 'zan')
        'amin': '',            # A min        (use with 'zanrange')
        'amax': '',            # A max        (use with 'zanrange')
        'n': '',               # N            (use with 'zan')
        'nmin': '',            # N min        (use with 'zanrange')
        'nmax': '',            # N max        (use with 'zanrange')
        'evenz': '',           # 'any', 'even', or 'odd' Z (use with zanrange)
        'evena': '',           # 'any', 'even', or 'odd' A (use with zanrange')
        'evenn': '',           # 'any', 'even', or 'odd' N (use with zanrange)
        'tled': 'disabled',    # half-life condition on/off
        'tlmin': '0',          # half-life min
        'utlow': 'S',          # half-life min units ('S' = seconds)
        'tlmax': '3E17',       # half-life max
        'utupp': 'S',          # half-life max units ('ST' = stable, 'GY' = Gy)
        'notlim': 'disabled',  # half-life: no limit
        'dmed': 'disabled',    # decay mode condition on/off
        'dmn': 'ANY',          # decay mode: 'ANY' = any
        'out': 'file',         # output to formatted file
        'unc': 'stdandard',    # standard style uncertainties
        'sub': 'Search',       # search for the data
    }
    _ALLOWED_KEYWORDS = [
        'perform', 'nuc', 'z', 'a', 'n',
        'z_range', 'a_range', 'n_range',
        'z_any', 'z_even', 'z_odd',
        'a_any', 'a_even', 'a_odd',
        'n_any', 'n_even', 'n_odd',
        't_range',
    ]
    _DUMMY_TEXT = ''

    def __init__(self, **kwargs):
        """Initialize query of NNDC data."""
        perform = kwargs.get('perform', True)
        kwargs['perform'] = False
        self._data = dict(self._DATA)
        self._text = self._DUMMY_TEXT
        self.df = pd.DataFrame()
        self.update(**kwargs)
        if perform:
            self.perform()

    def __len__(self):
        """Length of any one of the data lists."""
        if self.df is None:
            return 0
        elif len(self.df.keys()) == 0:
            return 0
        else:
            return len(self.df[self.df.keys()[0]])

    def keys(self):
        """Return the data keys."""
        return self.df.keys()

    def __getitem__(self, key):
        """Return the list given by the key."""
        return self.df[key]

    def __setitem__(self, key, value):
        """Set the list given by the key."""
        self.df[key] = value

    def __str__(self):
        """Use str method for DataFrame."""
        return str(self.df)

    def __format__(self, formatstr):
        """Use format method for DataFrame."""
        return self.df.__format__(formatstr)

    def _request(self):
        """Request data table from the URL."""
        # Use the context manager to automatically cleanup the session
        with requests.Session() as session:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=ResourceWarning)
                resp = session.post(self._URL, data=self._data, stream=False)
            if not resp.ok or resp.reason != 'OK' or resp.status_code != 200:
                raise NNDCRequestError('Request failed: ' + resp.reason)
            for msg in [
                    'Your search was unsuccessful',
                    'There are too many results for your search',
            ]:
                if msg in resp.text:
                    raise NNDCRequestError('Request failed: ' + msg)
            msg = 'No datasets were found within the specified search'
            if msg in resp.text:
                raise NoDataFound(msg)
        return resp.text

    def update(self, **kwargs):
        """Update the search criteria."""
        for kwarg in kwargs:
            if kwarg not in self._ALLOWED_KEYWORDS:
                raise NNDCInputError('Unknown keyword: "{}"'.format(kwarg))
        if 'nuc' in kwargs:
            self._data['spnuc'] = 'name'
            self._data['nuc'] = kwargs['nuc']
        for x in ['z', 'a', 'n']:
            # handle Z, A, and N settings
            if x in kwargs:
                self._data['spnuc'] = 'zanrange'
                self._data[x + 'min'], self._data[x + 'max'] = \
                    _format_range((kwargs[x], kwargs[x]))
            # handle *_range, *_any, *_odd, *_even
            elif x + '_range' in kwargs:
                self._data['spnuc'] = 'zanrange'
                self._data[x + 'min'], self._data[x + 'max'] = \
                    _format_range(kwargs[x + '_range'])
                if self._data[x + 'min'] == '':
                    self._data[x + 'min'] = '0'
                if self._data[x + 'max'] == '':
                    self._data[x + 'max'] = '300'
            if x + '_any' in kwargs:
                self._data['even' + x] = 'any'
            elif x + '_even' in kwargs:
                self._data['even' + x] = 'even'
            elif x + '_odd' in kwargs:
                self._data['even' + x] = 'odd'
        # handle half-life range condition
        if 't_range' in kwargs:
            self._data['tled'] = 'enabled'
            self._data['tlmin'], self._data['tlmax'] = \
                _format_range(kwargs['t_range'])

    def perform(self):
        """Perform the query."""
        # check the conditions
        if self._data['spnuc'] == '':
            self.update(z_range=(None, None))
        # submit the query
        try:
            self._text = self._request()
        except NoDataFound:
            self._text = self._DUMMY_TEXT
        if len(self._text) == 0:
            raise NNDCRequestError('NNDC returned no text')
        # package the output into a dictionary of arrays
        data = _parse_table(self._text)
        # create the DataFrame
        self.df = pd.DataFrame(data)
        # convert dimensionless integers to ints
        for col in ['A', 'Z', 'N', 'M']:
            if col in self.keys():
                self._convert_column(col, int)
        # combine uncertainty columns and add unit labels
        self._add_units_uncertainties()
        # add some more columns
        self._add_columns_energy_levels()
        # sort columns
        self._sort_columns()

    def _add_columns_energy_levels(self):
        """Add nuclear energy level 'M' and 'm' columns using energy levels."""
        if 'Energy Level (MeV)' not in self.df:
            return
        # add column of integer M giving the isomer level (0, 1, 2, ...)
        self.df['M'] = [0] * len(self)
        # add string m giving the isomer level name (e.g., '' or 'm' or 'm2')
        self.df['m'] = [''] * len(self)
        # loop over each isotope in the dataframe
        A_Z = [(a, z) for a, z in zip(self['A'], self['Z'])]
        A_Z = set(A_Z)
        for a, z in A_Z:
            isotope = (self['A'] == a) & (self['Z'] == z)
            e_levels = []
            e_levels_nominal = []
            for e_level in self['Energy Level (MeV)'][isotope]:
                if isinstance(e_level, uncertainties.core.Variable):
                    e_level_nominal = e_level.nominal_value
                else:
                    e_level_nominal = e_level
                if e_level_nominal not in e_levels_nominal:
                    e_levels.append(e_level)
                    e_levels_nominal.append(e_level_nominal)
            e_levels = sorted(e_levels)
            for M, e_level in enumerate(e_levels):
                isomer = isotope & \
                    (abs(self['Energy Level (MeV)'] - e_level) < 1e-10)
                self.df.loc[isomer, 'M'] = M
                if M > 0:
                    if len(e_levels) > 2:
                        self.df.loc[isomer, 'm'] = 'm{}'.format(M)
                    else:
                        self.df.loc[isomer, 'm'] = 'm'

    def _add_units_uncertainties(self):
        """Add units and uncertainties with some columns as applicable."""
        if 'Energy Level' in self.keys():
            self._convert_column(
                'Energy Level',
                lambda x: _parse_float_uncertainty(x, ''))
            self.df.rename(
                columns={'Energy Level': 'Energy Level (MeV)'}, inplace=True)

        if 'Parent Energy Level' in self.keys():
            self._convert_column_uncertainty('Parent Energy Level')
            self.df.rename(
                columns={'Parent Energy Level': 'Energy Level (MeV)'},
                inplace=True)
            self.df['Energy Level (MeV)'] *= 0.001

        if 'Mass Excess' in self.keys():
            self._convert_column_uncertainty('Mass Excess')
        self.df.rename(
            columns={'Mass Excess': 'Mass Excess (MeV)'}, inplace=True)

        self._convert_column('T1/2 (s)', float)

        if 'Abundance (%)' in self.keys():
            self._convert_column_uncertainty('Abundance (%)')

        if 'Branching (%)' in self.keys():
            self._convert_column(
                'Branching (%)',
                lambda x: _parse_float_uncertainty(x, ''))

        if 'Radiation Energy' in self.keys():
            self._convert_column_uncertainty('Radiation Energy')
            self.df.rename(
                columns={'Radiation Energy': 'Radiation Energy (keV)'},
                inplace=True)

        if 'Endpoint Energy' in self.keys():
            self._convert_column_uncertainty('Endpoint Energy')
            self.df.rename(
                columns={'Endpoint Energy': 'Endpoint Energy (keV)'},
                inplace=True)

        if 'Radiation Intensity (%)' in self.keys():
            self._convert_column_uncertainty('Radiation Intensity (%)')

        if 'Dose' in self.keys():
            self._convert_column_uncertainty('Dose')
            self.df.rename(
                columns={'Dose': 'Dose (MeV / Bq / s)'}, inplace=True)

    def _convert_column(self, col, function):
        """Convert column from string to another type."""
        col_new = []
        for x in self[col]:
            if x == '':
                col_new.append(None)
            else:
                col_new.append(function(x))
        self.df[col] = col_new

    def _convert_column_uncertainty(self, col):
        """Combine column and its uncertainty into one column."""
        col_new = []
        for x, dx in zip(self[col], self[col + ' Unc.']):
            x2 = _parse_float_uncertainty(x, dx)
            col_new.append(x2)
        self.df[col] = col_new
        del self.df[col + ' Unc.']

    def _sort_columns(self):
        """Sort columns."""
        preferred_order = [
            'Z', 'Element', 'A', 'm', 'M', 'N', 'JPi', 'T1/2',
            'Energy Level (MeV)', 'Decay Mode', 'Branching (%)',
            'Radiation', 'Radiation subtype',
            'Radiation Energy (keV)', 'Radiation Intensity (%)',
        ]
        new_cols = []
        for col in preferred_order:
            if col in self.keys():
                new_cols.append(col)
        for col in self.keys():
            if col not in new_cols:
                new_cols.append(col)
        self.df = self.df[new_cols]


class _NuclearWalletCardQuery(_NNDCQuery):
    """NNDC Nuclear Wallet Card data query.

    Nuclear Wallet Card Search can be performed at this URL:
        http://www.nndc.bnl.gov/nudat2/indx_sigma.jsp

    Help page: http://www.nndc.bnl.gov/nudat2/help/wchelp.jsp

      * Energy: Level energy in MeV.
      * JPi: Level spin and parity.
      * Mass Exc: Level Mass Excess in MeV.
      * T1/2 (txt): Level half-life in the format value+units+uncertainty.
      * T1/2 (seconds): value of the level half-life in seconds.
        Levels that are stable are assigned an "infinity" value.
      * Abund.: Natural abundance.
      * Dec Mode: Decay Mode name.
      * Branching (%): Percentual branching ratio for the corresponding
            decay mode.

    Args:
      perform: a boolean dictating whether to immediately perform the query.
      nuc     : (str) : the name of the isotope (e.g., 'Co-60')
      z, a, n : (int) : Z, A, N of the isotope
      z_range, etc. : (tuple of int) : range of Z, A, or N
      z_any, etc. : (bool) : whether any Z, A, or N is considered
      z_odd, etc. : (bool) : only odd Z, A, or N
      z_even, etc.: (bool) : only even Z, A, or N
      t_range : (tuple of float) : range of isotope half-lives in seconds
      elevel_range : (tuple of float) : range of nuc. energy level (MeV)
      decay : (str) : isotope decay mode from WALLET_DECAY_MODE
      j :  (str) : nuclear spin
      parity : (str) : nuclear parity

    Raises:
      NNDCInputError: if there is a problem with the input.
      NNDCRequestError: if there was a problem with the data requested.

    """

    _URL = 'https://www.nndc.bnl.gov/nudat2/sigma_searchi.jsp'
    _DATA = dict(_NNDCQuery._DATA)
    _DATA.update({
        'eled': 'disabled',    # E(level) condition on/off
        'elmin': '0',          # E(level) min
        'elmax': '40',         # E(level) max
        'jled': 'disabled',    # J_pi(level) condition on/off
        'jlv': '',             # J
        'plv': 'ANY',          # parity
        'ord': 'zalt',         # order file by Z, A, E(level), T1/2
    })
    _ALLOWED_KEYWORDS = list(_NNDCQuery._ALLOWED_KEYWORDS)
    _ALLOWED_KEYWORDS.extend(['elevel_range', 'decay', 'j', 'parity'])
    _DUMMY_TEXT = """
<html>
<body>

<pre>
A  	Element	Z  	N  	Energy  	JPi           	Mass Exc  	Unc  	T1/2 (txt)               	T1/2 (seconds)           	Abund.    	Unc     	Dec Mode	Branching (%)

 To save this output into a local file, clik on "File" in your browser menu and select "Save as"

</pre></body></html>
"""

    def update(self, **kwargs):
        """Update the search criteria."""
        super().update(**kwargs)
        # handle decay mode
        if 'decay' in kwargs:
            if kwargs['decay'].lower() not in WALLET_DECAY_MODE:
                raise NNDCInputError(
                    'Decay mode must be one of {}, not {}'.format(
                        WALLET_DECAY_MODE.keys(), kwargs['decay'].lower()))
            self._data['dmed'] = 'enabled'
            self._data['dmn'] = WALLET_DECAY_MODE[kwargs['decay'].lower()]
        # handle energy level condition
        if 'elevel_range' in kwargs:
            self._data['eled'] = 'enabled'
            self._data['elmin'], self._data['elmax'] = \
                _format_range(kwargs['elevel_range'])
            if self._data['elmax'] == '':
                self._data['elmax'] = '1000000000'
        # handle spin and parity
        if 'j' in kwargs:
            self._data['jled'] = 'enabled'
            self._data['jlv'] = kwargs['j']
        if 'parity' in kwargs:
            if kwargs['parity'].lower() not in PARITIES:
                raise NNDCInputError(
                    'Parity must be one of {}, not {}'.format(
                        PARITIES, kwargs['parity'].lower()))
            self._data['jled'] = 'enabled'
            self._data['plv'] = kwargs['parity'].upper()


def fetch_wallet_card(**kwargs):
    """Perform NNDC Nuclear Wallet Card data query and return a DataFrame.

    Nuclear Wallet Card Search can be performed at this URL:
        http://www.nndc.bnl.gov/nudat2/indx_sigma.jsp

    Help page: http://www.nndc.bnl.gov/nudat2/help/wchelp.jsp

      * Energy: Level energy in MeV.
      * JPi: Level spin and parity.
      * Mass Exc: Level Mass Excess in MeV.
      * T1/2 (txt): Level half-life in the format value+units+uncertainty.
      * T1/2 (seconds): value of the level half-life in seconds.
        Levels that are stable are assigned an "infinity" value.
      * Abund.: Natural abundance.
      * Dec Mode: Decay Mode name.
      * Branching (%): Percentual branching ratio for the corresponding
            decay mode.

    Args:
      nuc     : (str) : the name of the isotope (e.g., 'Co-60')
      z, a, n : (int) : Z, A, N of the isotope
      z_range, etc. : (tuple of int) : range of Z, A, or N
      z_any, etc. : (bool) : whether any Z, A, or N is considered
      z_odd, etc. : (bool) : only odd Z, A, or N
      z_even, etc.: (bool) : only even Z, A, or N
      t_range : (tuple of float) : range of isotope half-lives in seconds
      elevel_range : (tuple of float) : range of nuc. energy level (MeV)
      decay : (str) : isotope decay mode from WALLET_DECAY_MODE
      j :  (str) : nuclear spin
      parity : (str) : nuclear parity

    Returns:
      pandas DataFrame with the requested data.

    Raises:
      NNDCInputError: if there is a problem with the input.
      NNDCRequestError: if there was a problem with the data requested.

    """

    query = _NuclearWalletCardQuery(**kwargs)
    return query.df


class _DecayRadiationQuery(_NNDCQuery):
    """NNDC Decay Radiation data query.

    Decay Radiation Search can be performed at this URL:
        http://www.nndc.bnl.gov/nudat2/indx_dec.jsp

    Help page: http://www.nndc.bnl.gov/nudat2/help/dehelp.jsp

      * Radiation: Radiation type, i.e. G for gamma, E for electron.
      * Rad subtype: Further classification of the radiation type.
      * Rad Ene.: Radiation energy in keV.
      * EP Ene.: Beta-decay end point energy in keV.
      * Rad Int.: Radiation absolute intensity.
      * Dose: Radiation dose in MeV/Bq-s
      * Unc: Uncertainties

    Args:
      nuc     : (str) : the name of the isotope (e.g., 'Co-60')
      z, a, n : (int) : Z, A, N of the isotope
      z_range, etc. : (tuple of int) : range of Z, A, or N
      z_any, etc. : (bool) : whether any Z, A, or N is considered
      z_odd, etc. : (bool) : only odd Z, A, or N
      z_even, etc.: (bool) : only even Z, A, or N
      t_range : (tuple of float) : range of isotope half-lives in seconds
      decay : (str) : isotope decay mode from DECAYRAD_DECAY_MODE
      elevel_range : (tuple of float) : range of parent energy level (MeV)
      type :  (str) : radiation type from DECAYRAD_RADIATION_TYPE
      e_range : (tuple of float) : radiation energy range (keV)
      i_range : (tuple of float): intensity range (percent)

    Raises:
      NNDCInputError: if there is a problem with the input.
      NNDCRequestError: if there was a problem with the data requested.

    """

    _URL = 'https://www.nndc.bnl.gov/nudat2/dec_searchi.jsp'
    _DATA = dict(_NNDCQuery._DATA)
    _DATA.update({
        'rted': 'enabled',     # radiation type condition on/off
        'rtn': 'ANY',          # radiation type: 'ANY' = any, 'G' = gamma
        'reed': 'disabled',    # radiation energy condition on/off
        'remin': '0',          # radiation energy min (keV)
        'remax': '10000',      # radiation energy max (keV)
        'ried': 'disabled',    # radiation intensity condition on/off
        'rimin': '0',          # radiation intensity min (%)
        'rimax': '100',        # radiation intensity max (%)
        'ord': 'zate',         # order file by Z, A, T1/2, E
    })
    _ALLOWED_KEYWORDS = list(_NNDCQuery._ALLOWED_KEYWORDS)
    _ALLOWED_KEYWORDS.extend([
        'elevel_range', 'decay', 'type', 'e_range', 'i_range'])
    _DUMMY_TEXT = """
<html>
<body>
<pre>
A  	Element	Z  	N  	Par. Elevel	Unc. 	JPi       	Dec Mode	T1/2 (txt)    	T1/2 (num)        	Daughter	Radiation	Rad subtype 	Rad Ene.  	Unc       	EP Ene.   	Unc       	Rad Int.  	Unc       	Dose        	Unc

</pre>
To save this output into a local File, clik on "File" in your browser menu and select "Save as"
</body></html>
"""

    def update(self, **kwargs):
        """Update the search criteria."""
        super().update(**kwargs)
        # handle decay mode
        if 'decay' in kwargs:
            if kwargs['decay'].lower() not in DECAYRAD_DECAY_MODE:
                raise NNDCInputError(
                    'Decay mode must be one of {}, not {}'.format(
                        DECAYRAD_DECAY_MODE.keys(), kwargs['decay'].lower()))
            self._data['dmed'] = 'enabled'
            self._data['dmn'] = DECAYRAD_DECAY_MODE[kwargs['decay'].lower()]
        # handle radiation type
        if 'type' in kwargs:
            if kwargs['type'].lower() not in DECAYRAD_RADIATION_TYPE:
                raise NNDCInputError(
                    'Radiation type must be one of {}, not {}'.format(
                        DECAYRAD_RADIATION_TYPE.keys(),
                        kwargs['type'].lower()))
            self._data['rted'] = 'enabled'
            self._data['rtn'] = DECAYRAD_RADIATION_TYPE[kwargs['type'].lower()]
        # handle energy level condition
        self.elevel_range = (0, 1e9)
        if 'elevel_range' in kwargs:
            x = _format_range(kwargs['elevel_range'])
            try:
                x0 = float(x[0])
            except ValueError:
                x0 = 0.
            try:
                x1 = float(x[1])
            except ValueError:
                x1 = 1e9
            self.elevel_range = (x0, x1)
        # handle radiation energy range
        if 'e_range' in kwargs:
            self._data['reed'] = 'enabled'
            self._data['remin'], self._data['remax'] = \
                _format_range(kwargs['e_range'])
        # handle radiation intensity range
        if 'i_range' in kwargs:
            self._data['ried'] = 'enabled'
            self._data['rimin'], self._data['rimax'] = \
                _format_range(kwargs['i_range'])


def fetch_decay_radiation(**kwargs):
    """Perform NNDC Decay Radiation data query and return a DataFrame.

    Decay Radiation Search can be performed at this URL:
        http://www.nndc.bnl.gov/nudat2/indx_dec.jsp

    Help page: http://www.nndc.bnl.gov/nudat2/help/dehelp.jsp

      * Radiation: Radiation type, i.e. G for gamma, E for electron.
      * Rad subtype: Further classification of the radiation type.
      * Rad Ene.: Radiation energy in keV.
      * EP Ene.: Beta-decay end point energy in keV.
      * Rad Int.: Radiation absolute intensity.
      * Dose: Radiation dose in MeV/Bq-s
      * Unc: Uncertainties

    Args:
      nuc     : (str) : the name of the isotope (e.g., 'Co-60')
      z, a, n : (int) : Z, A, N of the isotope
      z_range, etc. : (tuple of int) : range of Z, A, or N
      z_any, etc. : (bool) : whether any Z, A, or N is considered
      z_odd, etc. : (bool) : only odd Z, A, or N
      z_even, etc.: (bool) : only even Z, A, or N
      t_range : (tuple of float) : range of isotope half-lives in seconds
      elevel_range : (tuple of float) : range of parent energy level (MeV)
      decay : (str) : isotope decay mode from DECAYRAD_DECAY_MODE
      type :  (str) : radiation type from DECAYRAD_RADIATION_TYPE
      e_range : (tuple of float) : radiation energy range (keV)
      i_range : (tuple of float): intensity range (percent)

    Returns:
      pandas DataFrame with the requested data.

    Raises:
      NNDCInputError: if there is a problem with the input.
      NNDCRequestError: if there was a problem with the data requested.

    """

    query = _DecayRadiationQuery(**kwargs)
    # apply elevel_range filter (hack around the web API)
    elevel = query.df['Energy Level (MeV)']
    keep = (elevel >= query.elevel_range[0]) & \
           (elevel <= query.elevel_range[1])
    query.df = query.df[keep]
    return query.df
